#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import List, Set, Dict
from ....lib import Person
from ....db import Database
from ....types import PersonHandle


# -------------------------------------------------------------------------
#
# RelationshipPathBetween
#
# -------------------------------------------------------------------------
class RelationshipPathBetween(Rule):
    """Rule that checks for a person that is a descendant of a specified person
    not more than N generations away"""

    labels = [_("ID:"), _("ID:")]
    name = _("Relationship path between <persons>")
    category = _("Relationship filters")
    description = _(
        "Matches the ancestors of two persons back "
        "to a common ancestor, producing the relationship "
        "path between two persons."
    )

    def prepare(self, db: Database, user):
        self.db = db
        self.selected_handles: Set[PersonHandle] = set()
        root1 = db.get_person_from_gramps_id(self.list[0])
        root2 = db.get_person_from_gramps_id(self.list[1])
        if root1 and root2:
            self.init_list(root1.handle, root2.handle)

    def reset(self):
        self.selected_handles.clear()

    def desc_list(self, handle: PersonHandle, map, first: bool):
        if not first:
            map.add(handle)

        p = self.db.get_person_from_handle(handle)
        for fam_id in p.family_list:
            fam = self.db.get_family_from_handle(fam_id)
            if fam:
                for child_ref in fam.child_ref_list:
                    if child_ref.ref:
                        self.desc_list(child_ref.ref, map, False)

    def apply_filter(
        self,
        rank: int,
        handle: PersonHandle,
        plist: Set[PersonHandle],
        pmap: Dict[PersonHandle, int],
    ):
        if not handle:
            return
        person = self.db.get_person_from_handle(handle)
        if person is None:
            return
        plist.add(handle)
        pmap[person.handle] = rank

        fam_id = (
            person.parent_family_list[0] if len(person.parent_family_list) > 0 else None
        )
        if not fam_id:
            return
        family = self.db.get_family_from_handle(fam_id)
        if family is not None:
            self.apply_filter(rank + 1, family.father_handle, plist, pmap)
            self.apply_filter(rank + 1, family.mother_handle, plist, pmap)

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles

    def init_list(self, p1_handle: PersonHandle, p2_handle: PersonHandle):
        firstMap: Dict[PersonHandle, int] = {}
        firstList: Set[PersonHandle] = set()
        secondMap: Dict[PersonHandle, int] = {}
        secondList: Set[PersonHandle] = set()
        common: List[PersonHandle] = []
        rank = 9999999

        self.apply_filter(0, p1_handle, firstList, firstMap)
        self.apply_filter(0, p2_handle, secondList, secondMap)

        for person_handle in firstList and secondList:
            new_rank = firstMap[person_handle]
            if new_rank < rank:
                rank = new_rank
                common = [person_handle]
            elif new_rank == rank:
                common.append(person_handle)

        path1 = set([p1_handle])
        path2 = set([p2_handle])

        for person_handle in common:
            new_map: Set[PersonHandle] = set()
            self.desc_list(person_handle, new_map, True)
            path1.update(new_map.intersection(firstMap))
            path2.update(new_map.intersection(secondMap))
        self.selected_handles.update(path1, path2, common)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import unittest

from gramps.gen.display.name import NameDisplay
from gramps.gen.lib import Name


class NameTest(unittest.TestCase):
    def setUp(self):
        self.name_display = NameDisplay()

    def test_get_name_format_for_all_without_default_format(self):
        self.add_custom_name_format("Surname, Name|Common Suffix")
        self.add_inactive_custom_name_format("SURNAME, Given Suffix (Call)")

        actual_name_format = self.name_display.get_name_format(
            also_default=False, only_custom=False, only_active=False
        )

        expected_name_format = [
            (1, "Surname, Given Suffix", "%l, %f %s", True),
            (2, "Given Surname Suffix", "%f %l %s", True),
            (3, "Patronymic, Given", "%y, %s %f", False),
            (4, "Given", "%f", True),
            (
                5,
                "Main Surnames, Given Patronymic Suffix Prefix",
                "%1m %2m %o, %f %1y %s %0m",
                True,
            ),
            (-2, "SURNAME, Given Suffix (Call)", "SURNAME, Given Suffix (Call)", False),
            (-1, "Surname, Name|Common Suffix", "Surname, Name|Common Suffix", True),
        ]
        self.assertEqual(expected_name_format, actual_name_format)

    def test_get_name_format_for_all_active_without_default_format(self):
        self.add_custom_name_format("Surname, Name|Common Suffix")
        self.add_inactive_custom_name_format("SURNAME, Given Suffix (Call)")

        actual_name_format = self.name_display.get_name_format(
            also_default=False, only_custom=False, only_active=True
        )

        expected_name_format = [
            (1, "Surname, Given Suffix", "%l, %f %s", True),
            (2, "Given Surname Suffix", "%f %l %s", True),
            (4, "Given", "%f", True),
            (
                5,
                "Main Surnames, Given Patronymic Suffix Prefix",
                "%1m %2m %o, %f %1y %s %0m",
                True,
            ),
            (-1, "Surname, Name|Common Suffix", "Surname, Name|Common Suffix", True),
        ]
        self.assertEqual(expected_name_format, actual_name_format)

    def test_get_name_format_for_all_custom_formats_without_default_format(self):
        self.add_custom_name_format("Surname, Name|Common Suffix")
        self.add_inactive_custom_name_format("SURNAME, Given Suffix (Call)")

        actual_name_format = self.name_display.get_name_format(
            also_default=False, only_custom=True, only_active=False
        )

        expected_name_format = [
            (-2, "SURNAME, Given Suffix (Call)", "SURNAME, Given Suffix (Call)", False),
            (-1, "Surname, Name|Common Suffix", "Surname, Name|Common Suffix", True),
        ]
        self.assertEqual(expected_name_format, actual_name_format)

    def test_get_name_format_for_active_custom_formats_without_default_format(self):
        self.add_custom_name_format("Surname, Name|Common Suffix")
        self.add_inactive_custom_name_format("SURNAME, Given Suffix (Call)")

        actual_name_format = self.name_display.get_name_format(
            also_default=False, only_custom=True, only_active=True
        )

        expected_name_format = [
            (-1, "Surname, Name|Common Suffix", "Surname, Name|Common Suffix", True)
        ]
        self.assertEqual(expected_name_format, actual_name_format)

    def test_get_name_format_for_all(self):
        self.add_custom_name_format("Surname, Name|Common Suffix")
        self.add_inactive_custom_name_format("SURNAME, Given Suffix (Call)")

        actual_name_format = self.name_display.get_name_format(
            also_default=True, only_custom=False, only_active=False
        )

        expected_name_format = [
            (0, "Default format (defined by Gramps preferences)", "", True),
            (1, "Surname, Given Suffix", "%l, %f %s", True),
            (2, "Given Surname Suffix", "%f %l %s", True),
            (3, "Patronymic, Given", "%y, %s %f", False),
            (4, "Given", "%f", True),
            (
                5,
                "Main Surnames, Given Patronymic Suffix Prefix",
                "%1m %2m %o, %f %1y %s %0m",
                True,
            ),
            (-2, "SURNAME, Given Suffix (Call)", "SURNAME, Given Suffix (Call)", False),
            (-1, "Surname, Name|Common Suffix", "Surname, Name|Common Suffix", True),
        ]
        self.assertEqual(expected_name_format, actual_name_format)

    def test_get_name_format_for_all_active(self):
        self.add_custom_name_format("Surname, Name|Common Suffix")
        self.add_inactive_custom_name_format("SURNAME, Given Suffix (Call)")

        actual_name_format = self.name_display.get_name_format(
            also_default=True, only_custom=False, only_active=True
        )

        expected_name_format = [
            (0, "Default format (defined by Gramps preferences)", "", True),
            (1, "Surname, Given Suffix", "%l, %f %s", True),
            (2, "Given Surname Suffix", "%f %l %s", True),
            (4, "Given", "%f", True),
            (
                5,
                "Main Surnames, Given Patronymic Suffix Prefix",
                "%1m %2m %o, %f %1y %s %0m",
                True,
            ),
            (-1, "Surname, Name|Common Suffix", "Surname, Name|Common Suffix", True),
        ]
        self.assertEqual(expected_name_format, actual_name_format)

    def test_get_name_format_for_all_custom_formats(self):
        self.add_custom_name_format("Surname, Name|Common Suffix")
        self.add_inactive_custom_name_format("SURNAME, Given Suffix (Call)")

        actual_name_format = self.name_display.get_name_format(
            also_default=True, only_custom=True, only_active=False
        )

        expected_name_format = [
            (-2, "SURNAME, Given Suffix (Call)", "SURNAME, Given Suffix (Call)", False),
            (-1, "Surname, Name|Common Suffix", "Surname, Name|Common Suffix", True),
        ]
        self.assertEqual(expected_name_format, actual_name_format)

    def test_get_name_format_for_active_custom_formats(self):
        self.add_custom_name_format("Surname, Name|Common Suffix")
        self.add_inactive_custom_name_format("SURNAME, Given Suffix (Call)")

        actual_name_format = self.name_display.get_name_format(
            also_default=True, only_custom=True, only_active=True
        )

        expected_name_format = [
            (-1, "Surname, Name|Common Suffix", "Surname, Name|Common Suffix", True)
        ]
        self.assertEqual(expected_name_format, actual_name_format)

    def test_clear_custom_formats(self):
        self.add_custom_name_format("Surname, Name|Common Suffix")
        self.add_inactive_custom_name_format("SURNAME, Given Suffix (Call)")

        self.name_display.clear_custom_formats()

        actual_name_format = self.name_display.get_name_format(
            also_default=False, only_custom=True, only_active=False
        )

        self.assertEqual([], actual_name_format)

    def test_do_not_clear_default_formats(self):
        self.add_custom_name_format("Surname, Name|Common Suffix")
        self.add_inactive_custom_name_format("SURNAME, Given Suffix (Call)")

        self.name_display.clear_custom_formats()

        actual_name_format = self.name_display.get_name_format(
            also_default=True, only_custom=False, only_active=False
        )

        expected_name_format = [
            (0, "Default format (defined by Gramps preferences)", "", True),
            (1, "Surname, Given Suffix", "%l, %f %s", True),
            (2, "Given Surname Suffix", "%f %l %s", True),
            (3, "Patronymic, Given", "%y, %s %f", False),
            (4, "Given", "%f", True),
            (
                5,
                "Main Surnames, Given Patronymic Suffix Prefix",
                "%1m %2m %o, %f %1y %s %0m",
                True,
            ),
        ]

        self.assertEqual(expected_name_format, actual_name_format)

    def test_set_name_format(self):
        standard_formats_overrides = [
            (0, "Default", "", True),
            (1, "Surname", "%l", True),
            (2, "Suffix", "%s", True),
            (3, "Given", "%s", False),
            (4, "Patronymic", "%y", True),
            (5, "Prefix", "%0m", True),
        ]

        self.name_display.set_name_format(standard_formats_overrides)

        actual_name_format = self.name_display.get_name_format(
            also_default=True, only_custom=False, only_active=False
        )

        self.assertEqual(standard_formats_overrides, actual_name_format)

    def test_add_existing_name_format(self):
        self.assertEqual(
            self.add_custom_name_format("Surname, Name|Common Suffix"),
            self.add_custom_name_format("Surname, Name|Common Suffix"),
        )

    def test_delete_name_format(self):
        index = self.add_custom_name_format("Surname, Name|Common Suffix")
        self.name_display.del_name_format(index)

        actual_name_format = self.name_display.get_name_format(
            also_default=True, only_custom=True, only_active=True
        )

        self.assertEqual([], actual_name_format)

    def test_set_default_format_to_custom_format(self):
        index = self.add_custom_name_format("Surname, Name|Common Suffix")
        self.name_display.set_default_format(index)
        self.assertEqual(index, self.name_display.get_default_format())

    def test_set_default_format_to_non_existing_format(self):
        index = self.name_display.get_default_format()
        self.name_display.set_default_format(-100)
        self.assertEqual(index, self.name_display.get_default_format())

    def test_display_name_with_valid_format(self):
        name = Name()
        name.set_first_name("William")
        name.set_call_name("Will")
        name.set_display_as(4)

        actual_display_name = self.name_display.display_name(name)

        self.assertEqual("William", actual_display_name)

    def test_display_name_with_invalid_format(self):
        name = Name()
        name.set_first_name("William")
        name.set_call_name("Will")
        name.set_display_as(-100)

        display_name = self.name_display.display_name(name)

        self.assertEqual(", William", display_name)

    def test_display_no_name(self):
        self.assertEqual("", self.name_display.display_name(None))

    def add_custom_name_format(self, name_format):
        return self.name_display.add_name_format(name_format, name_format)

    def add_inactive_custom_name_format(self, name_format):
        index = self.name_display.add_name_format(name_format, name_format)
        self.name_display.set_format_inactive(index)


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Address class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .citationbase import CitationBase
from .const import DIFFERENT, EQUAL, IDENTICAL
from .datebase import DateBase
from .locationbase import LocationBase
from .notebase import NoteBase
from .privacybase import PrivacyBase
from .secondaryobj import SecondaryObject

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Address
#
# -------------------------------------------------------------------------
class Address(
    SecondaryObject, PrivacyBase, CitationBase, NoteBase, DateBase, LocationBase
):
    """
    Provides address information.
    """

    def __init__(self, source=None):
        """
        Create a new Address instance, copying from the source if provided.
        """
        PrivacyBase.__init__(self, source)
        CitationBase.__init__(self, source)
        NoteBase.__init__(self, source)
        DateBase.__init__(self, source)
        LocationBase.__init__(self, source)

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            PrivacyBase.serialize(self),
            CitationBase.serialize(self),
            NoteBase.serialize(self),
            DateBase.serialize(self),
            LocationBase.serialize(self),
        )

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (privacy, citation_list, note_list, date, location) = data

        PrivacyBase.unserialize(self, privacy)
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        DateBase.unserialize(self, date)
        LocationBase.unserialize(self, location)
        return self

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        # pylint: disable=import-outside-toplevel
        from .date import Date

        return {
            "type": "object",
            "title": _("Address"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "private": {"type": "boolean", "title": _("Private")},
                "citation_list": {
                    "type": "array",
                    "title": _("Citations"),
                    "items": {"type": "string", "maxLength": 50},
                },
                "note_list": {
                    "type": "array",
                    "title": _("Notes"),
                    "items": {"type": "string", "maxLength": 50},
                },
                "date": Date.get_schema(),
                "street": {"type": "string", "title": _("Street")},
                "locality": {"type": "string", "title": _("Locality")},
                "city": {"type": "string", "title": _("City")},
                "county": {"type": "string", "title": _("County")},
                "state": {"type": "string", "title": _("State")},
                "country": {"type": "string", "title": _("Country")},
                "postal": {"type": "string", "title": _("Postal Code")},
                "phone": {"type": "string", "title": _("Phone")},
            },
        }

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return LocationBase.get_text_data_list(self)

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return []

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return []

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return []

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        return (
            self.get_referenced_note_handles() + self.get_referenced_citation_handles()
        )

    def is_equivalent(self, other):
        """
        Return if this address is equivalent, that is agrees in location and
        date, to other.

        :param other: The address to compare this one to.
        :type other: Address
        :returns: Constant indicating degree of equivalence.
        :rtype: int
        """
        if (
            self.get_text_data_list() != other.get_text_data_list()
            or self.get_date_object() != other.get_date_object()
        ):
            return DIFFERENT
        if self.is_equal(other):
            return IDENTICAL
        return EQUAL

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this address.

        Lost: date, street, city, county, state, country, postal and phone of
        acquisition.

        :param acquisition: The address to merge with the present address.
        :type acquisition: Address
        """
        self._merge_privacy(acquisition)
        self._merge_note_list(acquisition)
        self._merge_citation_list(acquisition)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2018      Paul Culley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
A replacement UIManager and ActionGroup.
"""

import copy
import sys
import logging
import xml.etree.ElementTree as ET

from gi.repository import GLib, Gio, Gtk

from ..gen.const import GRAMPS_LOCALE as glocale
from ..gen.config import config


LOG = logging.getLogger("gui.uimanager")


ACTION_NAME = 0  # tuple index for action name
ACTION_CB = 1  # tuple index for action callback
ACTION_ACC = 2  # tuple index for action accelerator
ACTION_ST = 3  # tuple index for action state


class ActionGroup:
    """This class represents a group of actions that con be manipulated
    together.
    """

    def __init__(self, name, actionlist=None, prefix="win"):
        """
        @param name: the action group name, used to match to the 'groups'
                     attribute in the ui xml.
        @type name: string
        @type actionlist: list
        @param actionlist: the list of actions to add
            The list contains tuples with the following contents:
            string: Action Name
            method: signal callback function.
                    None if just adding an accelerator
            string: accelerator ex: '<Primary>Enter' or '' for no accelerator.
                    optional for non-stateful actions.
            state: initial state for stateful actions.
                'True' or 'False': the action is interpreted as a checkbox.
                'None': non stateful action (optional)
                'string': the action is interpreted as a Radio button
        @type prefix: str
        @param prefix: the prefix used by this group.  If not provided, 'win'
                       is assumed.
        """
        self.name = name
        self.actionlist = actionlist if actionlist else []
        self.prefix = prefix + "."
        self.act_group = None
        self.sensitive = True

    def add_actions(self, actionlist):
        """Add a list of actions to the current list
        @type actionlist: list
        @param actionlist: the list of actions to add
        """
        self.actionlist.extend(actionlist)


class UIManager:
    """
    This is Gramps UIManager, it is designed to replace the deprecated Gtk
    UIManager.  The replacement is not exact, but performs similar
    functions, in some case with the same method names and parameters.
    It is designed to be a singleton.  The menu portion of this is responsible
    only for Gramps main window menus and toolbar.
    This was implemented to extend Gtk.Builder functions to allow editing
    (merging) of the original builder XML with additional XML fragments during
    operations.  This allows changing of the menus and toolbar when the tree is
    loaded, views are changed etc.

    The ActionGroup portions can also be used by other windows.  Other windows
    needing menus or toolbars can create them via Gtk.Builder.
    """

    def __init__(self, app, initial_xml):
        """
        @param app: Gramps Gtk.Application reference
        @type app: Gtk.Application
        @param initial_xml: Initial (primary) XML string for Gramps menus and
            toolbar
        @type changexml: string

        The xml is basically Gtk Builder xml, in particular the various menu
        and toolbar elements.  It is possible to add other elements as well.
        The xml this supports has been extended in two ways;
        1) there is an added "groups=" attribute to elements.  This
           attribute associates the element with one or more named ActionGroups
           for making the element visible or not.  If 'groups' is missing, the
           element will be shown as long as enclosing elements are shown.  The
           element will be shown if the group is present and considered visible
           by the uimanager.  If more than one group is needed, they should be
           separated by a space.
        2) there is an added <placeholder> tag supported; this is used to mark
           a place where merged UI XML can be inserted.  During the update_menu
           processing, elements enclosed in this tag pair are promoted to the
           level of the placeholder tag, and the placeholder tag is removed.

        Note that any elements can be merged (replaced) by the
        add_ui_from_string method, not just placeholders.  This works by
        matching the "id=" attribute on the element, and replacing the
        original element with the one from the add method.

        Note that when added elements are removed by the remove_ui method, they
        are replaced by the containing xml (with the 'id=') only, so don't put
        anything inside of the containing xml to start with as it will be lost
        during editing.
        """
        self.app = app  # the Gtk.Application of Gramps
        self.et_xml = ET.fromstring(initial_xml)
        self.builder = None
        self.toolbar = None
        self.action_groups = []  # current list of action groups
        self.show_groups = []  # groups to show at the moment
        self.accel_dict = {}  # used to store accel overrides from file

    def update_menu(self, init=False):
        """This updates the menus and toolbar when there is a change in the
        ui; any addition or removal or set_visible operation needs to call
        this.  It is best to make the call only once, at the end, if multiple
        changes are to be made.
        It starts with the ET xml stored in self, cleans it up to meet the
        Gtk.Builder specifications, and then updates the ui.

        @param init: When True, this is first call and we set the builder
                     toolbar and menu to the application.
                     When False, we update the menus and toolbar
        @type init: bool
        """

        def iterator(parents):
            """This recursively goes through the ET xml and deals with the
            'groups' attribute and <placeholder> tags, which are not valid for
            builder.  Empty submenus are also removed.
            Groups processing removes elements that are not shown, as well as
            the 'groups' attribute itself.
            <placeholder> tags are removed and their enclosed elements are
            promoted to the level of the placeholder.

            @param parents: the current element to recursively process
            @type parents: ET element
            """
            indx = 0
            while indx < len(parents):
                child = parents[indx]
                if len(child) >= 1:
                    # Recurse until we have a stand-alone child
                    iterator(child)
                if (len(child) == 1 and child.tag == "submenu") or (
                    len(child) == 0 and child.tag == "section"
                ):
                    # remove empty submenus and sections
                    LOG.debug(("del", child.tag, child.attrib))
                    del parents[indx]
                    continue
                LOG.debug((child.attrib))
                groups = child.get("groups")
                if not groups:
                    indx += 1
                    continue
                del child.attrib["groups"]
                for group in groups.split(" "):
                    if group in self.show_groups:
                        indx += 1
                        break
                    else:
                        LOG.debug(
                            (
                                "del",
                                child.tag,
                                child.attrib,
                                parents.tag,
                                parents.attrib,
                            )
                        )
                        del parents[indx]
                        break
            # The following looks for 'placeholder' elements and if found,
            # promotes any children to the same level as the placeholder.
            # this allows the user to insert elements without using a section.
            indx = 0
            while indx < len(parents):
                if parents[indx].tag == "placeholder":
                    subtree = parents[indx]
                    LOG.debug(
                        (
                            "placholder del",
                            parents[indx].tag,
                            parents[indx].attrib,
                            parents.tag,
                            parents.attrib,
                        )
                    )
                    del parents[indx]
                    for child in subtree:
                        parents.insert(indx, child)
                        indx += 1
                else:
                    indx += 1

        if self.builder:
            toolbar = self.builder.get_object("ToolBar")  # previous toolbar

        # need to copy the tree so we can preserve original for later edits.
        editable = copy.deepcopy(self.et_xml)
        iterator(editable)  # clean up tree to builder specifications
        del iterator  # Needed for garbage collection
        # The following should work, but seems to have a Gtk bug
        # xml_str = ET.tostring(editable, encoding="unicode")

        xml_str = ET.tostring(editable).decode(encoding="ascii")

        # debugging
        # with open('try.xml', 'w', encoding='utf8') as file:
        #     file.write(xml_str)
        # with open('try.xml', encoding='utf8') as file:
        #     xml_str = file.read()
        # LOG.info(xml_str)
        self.builder = Gtk.Builder()
        self.builder.set_translation_domain(glocale.get_localedomain())
        self.builder.add_from_string(xml_str)
        if init:
            self.app.menubar = self.builder.get_object("menubar")
            self.app.set_menubar(self.app.menubar)
            return
        # The following is the only way I have found to update the menus.
        # app.set_menubar can apparently only be used once, before
        # ApplicationWindow creation, further uses do NOT cause the menus to
        # update.
        self.app.menubar.remove_all()
        section = self.builder.get_object("menubar-update")
        self.app.menubar.append_section(None, section)

        # the following updates the toolbar from the new builder
        toolbar_parent = toolbar.get_parent()
        tb_show = toolbar.get_visible()
        toolbar_parent.remove(toolbar)
        toolbar = self.builder.get_object("ToolBar")  # new toolbar
        toolbar_style = config.get("interface.toolbar-style")
        if toolbar_style == 0:
            toolbar.set_style(Gtk.ToolbarStyle.BOTH)
        elif toolbar_style == 1:
            toolbar.set_style(Gtk.ToolbarStyle.TEXT)
        else:
            toolbar.set_style(Gtk.ToolbarStyle.ICONS)
        toolbar_parent.pack_start(toolbar, False, True, 0)
        if tb_show:
            toolbar.show_all()
        else:
            toolbar.hide()
        LOG.info("*** Update ui")

    def add_ui_from_string(self, changexml):
        """This performs a merge operation on the xml elements that have
        matching 'id's between the current ui xml and change xml strings.
        The 'changexml' is a list of xml fragment strings used to replace
        matching elements in the current xml.

        There MUST one and only one matching id in the orig xml.
        @param changexml: list of xml fragments to merge into main
        @type changexml: list
        @return: changexml
        """
        try:
            for xml in changexml:
                if not xml:
                    # allow an xml fragment to be an empty string
                    continue
                update = ET.fromstring(xml)
                el_id = update.attrib["id"]
                # find the parent of the id'd element in original xml
                parent = self.et_xml.find(".//*[@id='%s'].." % el_id)
                if parent is not None:
                    # we found it, now delete original, inset updated
                    for indx in range(len(parent)):
                        if parent[indx].get("id") == el_id:
                            del parent[indx]
                            parent.insert(indx, update)
                else:
                    # updated item not present in original, just add it
                    # This allow addition of popups etc.
                    self.et_xml.append(update)
            # results = ET.tostring(self.et_xml, encoding="unicode")
            # LOG.info(results)
            LOG.info("*** Add ui")
            return changexml
        except:
            # the following is only here to assist debug
            LOG.debug("*****", sys.exc_info())
            LOG.debug(xml)
            LOG.debug(changexml)
            assert False

    def remove_ui(self, change_xml):
        """This removes the 'change_xml' from the current ui xml.  It works on
        any element with matching 'id', the actual element remains but any
        children are removed.
        The 'change_xml' is a list of xml strings originally used to replace
        matching elements in the current ui xml.
        @param change_xml: list of xml fragments to remove from main
        @type change_xml: list
        """
        #         if not change_xml:
        #             import pydevd
        #             pydevd.settrace()
        for xml in change_xml:
            if not xml:
                continue
            update = ET.fromstring(xml)
            el_id = update.attrib["id"]
            # find parent of id'd element
            element = self.et_xml.find(".//*[@id='%s']" % el_id)
            if element is not None:  # element may have already been deleted
                for dummy in range(len(element)):
                    del element[0]
        # results = ET.tostring(self.et_xml, encoding="unicode")
        # LOG.info(results)
        LOG.info("*** Remove ui")
        return

    def get_widget(self, obj):
        """Get the object from the builder.
        @param obj: the widget to get
        @type obj: string
        @return: the object
        """
        return self.builder.get_object(obj)

    def insert_action_group(self, group, gio_group=None):
        """
        This inserts (actually overwrites any matching actions) the action
        group's actions to the app.
        By default (with no gio_group), the action group is added to the main
        Gramps window and the group assumes a 'win' prefix.
        If not using the main window, the window MUST have the 'application'
        property set for the accels to work.  In this case the actiongroups
        must be created like the following:

            # create Gramps ActionGroup
            self.action_group = ActionGroup('name', actions, 'prefix')
            # create Gio action group
            act_grp = SimpleActionGroup()
            # associate window with Gio group and its prefix
            window.insert_action_group('prefix', act_grp)
            # make the window 'application' aware
            window.set_application(uimanager.app)
            # tell the uimanager about the groups.
            uimanager.insert_action_group(self.action_group, act_grp)

        @param group: the action group
        @type group: ActionGroup
        @param gio_group: the Gio action group associated with a window.
        @type gio_group: Gio.SimpleActionGroup
        """
        try:
            assert isinstance(group.actionlist, list)
            if gio_group:
                window_group = group.act_group = gio_group
            elif group.act_group:
                window_group = group.act_group
            else:
                window_group = group.act_group = self.app.window
            for item in group.actionlist:
                if not Gio.action_name_is_valid(item[ACTION_NAME]):
                    LOG.warning("**Invalid action name %s", item[ACTION_NAME])
                # deal with accelerator overrides from a file
                accel = self.accel_dict.get(group.prefix + item[ACTION_NAME])
                if accel:
                    self.app.set_accels_for_action(
                        group.prefix + item[ACTION_NAME], [accel]
                    )
                elif len(item) > 2 and item[ACTION_ACC]:
                    # deal with accelerators defined in the group
                    accels = self.app.get_actions_for_accel(item[ACTION_ACC])
                    if accels:
                        # diagnostic printout; a duplicate accelerator may be
                        # a problem if both are valid for the same window at
                        # the same time.  If the actions are for a different
                        # window, this is not an error.  Here we assume a
                        # different prefix is used for different windows.
                        for accel in accels:
                            if group.prefix in accel:
                                LOG.warning(
                                    "**Duplicate Accelerator %s", item[ACTION_ACC]
                                )
                    self.app.set_accels_for_action(
                        group.prefix + item[ACTION_NAME], [item[ACTION_ACC]]
                    )
                if len(item) <= 3:
                    # Normal stateless actions
                    action = Gio.SimpleAction.new(item[ACTION_NAME], None)
                    if item[ACTION_CB]:  # in case we have only accelerator
                        action.connect("activate", item[ACTION_CB])
                elif isinstance(item[ACTION_ST], str):
                    # Radio Actions
                    action = Gio.SimpleAction.new_stateful(
                        item[ACTION_NAME],
                        GLib.VariantType.new("s"),
                        GLib.Variant("s", item[ACTION_ST]),
                    )
                    action.connect("change-state", item[ACTION_CB])
                elif isinstance(item[ACTION_ST], bool):
                    # Checkbox actions
                    action = Gio.SimpleAction.new_stateful(
                        item[ACTION_NAME],
                        None,
                        GLib.Variant.new_boolean(item[ACTION_ST]),
                    )
                    action.connect("change-state", item[ACTION_CB])
                window_group.add_action(action)
            self.action_groups.append(group)
            # if action sensitivity was set prior to actually inserting into
            # UIManager, we need to do it now that we have the action
            if not group.sensitive:
                self.set_actions_sensitive(group, False)
        except:
            # the following is only to assist in debug
            LOG.debug(group.name, item)
            assert False

    def remove_action_group(self, group):
        """This removes the ActionGroup from the UIManager

        @param group: the action group
        @type group: ActionGroup
        """
        if group.act_group:
            window_group = group.act_group
        else:
            window_group = self.app.window
        for item in group.actionlist:
            window_group.remove_action(item[ACTION_NAME])
            self.app.set_accels_for_action(group.prefix + item[ACTION_NAME], [])
        self.action_groups.remove(group)

    def get_action_groups(self):
        """This returns a list of action Groups installed into the UIManager.
        @return: list of groups
        """
        return self.action_groups

    def set_actions_sensitive(self, group, value):
        """This sets an ActionGroup enabled or disabled.  A disabled action
        will be greyed out in the UI.

        @param group: the action group
        @type group: ActionGroup
        @param value: the state of the group
        @type value: bool
        """
        if group.act_group:
            for item in group.actionlist:
                action = group.act_group.lookup_action(item[ACTION_NAME])
                if action:
                    # We check in case the group has not been inserted into
                    # UIManager yet
                    action.set_enabled(value)
        group.sensitive = value

    def get_actions_sensitive(self, group):
        """This gets an ActionGroup sensitive setting.  A disabled action
        will be greyed out in the UI.
        We assume that the first action represents the group.

        @param group: the action group
        @type group: ActionGroup
        @return: the state of the group
        """
        item = group.actionlist[0]
        action = group.act_group.lookup_action(item[ACTION_NAME])
        return action.get_enabled()

    def set_actions_visible(self, group, value):
        """This sets an ActionGroup visible and enabled or invisible and
        disabled.  Make sure that the menuitems or sections and toolbar items
        have the 'groups=' xml attribute matching the group name for this to
        work correctly.

        @param group: the action group
        @type group: ActionGroup
        @param value: the state of the group
        @type value: bool
        """
        self.set_actions_sensitive(group, value)
        if value:
            if group.name not in self.show_groups:
                self.show_groups.append(group.name)
        else:
            if group.name in self.show_groups:
                self.show_groups.remove(group.name)

    def get_action(self, group, actionname):
        """Return a single action from the group.
        @param group: the action group
        @type group: ActionGroup
        @param actionname: the action name
        @type actionname: string
        @return: Gio.Action
        """
        return group.act_group.lookup_action(actionname)

    def enable_all_actions(self, state):
        for group in self.action_groups:
            if group.act_group:
                for item in group.actionlist:
                    action = group.act_group.lookup_action(item[ACTION_NAME])
                    if action:
                        # We check in case the group has not been inserted into
                        # UIManager yet
                        action.set_enabled(group.sensitive if state else False)

    def dump_all_accels(self):
        """A function used diagnostically to see what accels are present.
        This will only dump the current accel set, if other non-open windows
        or views have accels, you will need to open them and run this again
        and manually merge the result files.  The results are in a
        'gramps.accel' file located in the current working directory."""
        out_dict = {}
        for group in self.action_groups:
            for item in group.actionlist:
                act = group.prefix + item[ACTION_NAME]
                accels = self.app.get_accels_for_action(
                    group.prefix + item[ACTION_NAME]
                )
                out_dict[act] = accels[0] if accels else ""
        import json

        with open(
            "gramps.accel",
            "w",
        ) as hndl:
            accels = json.dumps(out_dict, indent=0).replace('\n"', '\n# "')
            hndl.write(accels)

    def load_accels(self, filename):
        """This function loads accels from a file such as created by
        dump_all_accels.  The file contents is basically a Python dict
        definition.  As such it contains a line for each dict element.
        These elements can be commented out with '#' at the beginning of the
        line.

        If used, this file overrides the accels defined in other Gramps code.
        As such it must be loaded before any insert_action_group calls.
        """
        import ast

        with open(filename, "r") as hndl:
            accels = hndl.read()
            self.accel_dict = ast.literal_eval(accels)


INVALID_CHARS = [" ", "_", "(", ")", ",", "'"]


def valid_action_name(text):
    """This function cleans up action names to avoid some illegal
    characters.  It does NOT clean up non-ASCII characters.
    This is used for plugin IDs to clean them up.  It would be better if we
    made all plugin ids:
    ASCII Alphanumeric and the '.' or '-' characters."""
    for char in INVALID_CHARS:
        text = text.replace(char, "-")
    return text



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008-2012  Brian G. Matherly
# Copyright (C) 2009       Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2012-2014  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
import datetime, time
import calendar

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.const import URL_HOMEPAGE
from gramps.gen.errors import ReportError
from gramps.gen.config import config
from gramps.gen.lib import NameType, EventType, Name, Date, Person, Surname
from gramps.gen.lib.date import gregorian, Today
from gramps.gen.relationship import get_relationship_calculator
from gramps.gen.plug.docgen import (
    FontStyle,
    ParagraphStyle,
    GraphicsStyle,
    FONT_SERIF,
    PARA_ALIGN_RIGHT,
    PARA_ALIGN_LEFT,
    PARA_ALIGN_CENTER,
    IndexMark,
    INDEX_TYPE_TOC,
)
from gramps.gen.plug.menu import (
    BooleanOption,
    StringOption,
    NumberOption,
    EnumeratedListOption,
    FilterOption,
    PersonOption,
)
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.utils.alive import probably_alive

import gramps.plugins.lib.libholiday as libholiday

# localization for BirthdayOptions only!!
from gramps.gen.datehandler import displayer as date_displayer


# _T_ is a gramps-defined keyword -- see po/update_po.py and po/genpot.sh
def _T_(value, context=""):  # enable deferred translations
    return "%s\x04%s" % (context, value) if context else value


_TITLE0 = _T_("Birthday and Anniversary Report")
_TITLE1 = _T_("My Birthday Report")
_TITLE2 = _T_("Produced with Gramps")
_DEADTXT = _T_("✝")


# ------------------------------------------------------------------------
#
# BirthdayReport
#
# ------------------------------------------------------------------------
class BirthdayReport(Report):
    """
    Create the BirthdayReport object that produces the report.

    name_format   - Preferred format to display names
    incl_private  - Whether to include private data
    """

    def __init__(self, database, options, user):
        Report.__init__(self, database, options, user)
        self._user = user
        menu = options.menu
        mgobn = lambda name: options.menu.get_option_by_name(name).get_value()

        stdoptions.run_private_data_option(self, menu)
        # (this report has its own "living people" option ("alive") already)

        self.titletext = mgobn("titletext")
        self.relationships = mgobn("relationships")
        self.year = mgobn("year")
        self.country = mgobn("country")
        self.maiden_name = mgobn("maiden_name")
        self.alive = mgobn("alive")
        self.birthdays = mgobn("birthdays")
        self.anniversaries = mgobn("anniversaries")
        self.death_anniversaries = mgobn("death_anniversaries")
        self.text1 = mgobn("text1")
        self.text2 = mgobn("text2")
        self.text3 = mgobn("text3")
        self.deadtxt = mgobn("deadtxt")
        self.filter_option = menu.get_option_by_name("filter")
        self.filter = self.filter_option.get_filter()
        self.showyear = mgobn("showyear")
        self.inc_gid = mgobn("inc_id")
        pid = mgobn("pid")

        self.set_locale(menu.get_option_by_name("trans").get_value())

        stdoptions.run_name_format_option(self, menu)

        self.center_person = self.database.get_person_from_gramps_id(pid)
        if self.center_person is None:
            raise ReportError(_("Person %s is not in the Database") % pid)

    def get_name(self, person, maiden_name=None):
        """
        Return person's name, unless maiden_name given, unless married_name
        listed.
        """
        # Get all of a person's names:
        primary_name = person.get_primary_name()
        married_name = None
        names = [primary_name] + person.get_alternate_names()
        for name in names:
            if int(name.get_type()) == NameType.MARRIED:
                married_name = name
                break  # use first
        # Now, decide which to use:
        if maiden_name is not None:
            if married_name is not None:
                name = Name(married_name)
            else:
                name = Name(primary_name)
                surname_obj = name.get_primary_surname()
                surname_obj.set_surname(maiden_name)
        else:
            name = Name(primary_name)
        if self.inc_gid:
            return "%s (%s)" % (
                self._name_display.display_name(name),
                person.get_gramps_id(),
            )
        else:
            return self._name_display.display_name(name)

    def add_day_item(self, text, month, day, person=None):
        """Add an item to a day."""
        month_dict = self.calendar.get(month, {})
        day_list = month_dict.get(day, [])
        day_list.append((text, person))
        month_dict[day] = day_list
        self.calendar[month] = month_dict

    def __get_holidays(self):
        """Get the holidays for the specified country and year"""
        holiday_table = libholiday.HolidayTable()
        country = holiday_table.get_countries()[self.country]
        holiday_table.load_holidays(self.year, country)
        for month in range(1, 13):
            for day in range(1, 32):
                holiday_names = holiday_table.get_holidays(month, day)
                for holiday_name in holiday_names:
                    self.add_day_item(self._(holiday_name), month, day)
                    # FIXME translation only works for a limited set of things
                    # (the right fix is to somehow feed the locale into the
                    # HolidayTable class in plugins/lib/libholiday.py and then
                    # probably changing all the holiday code to somehow defer
                    # the translation of holidays, until it can be based
                    # on the passed-in locale, but since that would probably
                    # also mean checking every use of holidays I don't think
                    # it is advisable to do, with a release so imminent)
                    # it is also debatable whether it is worth bothering at
                    # all, since it is hard for me to imagine why a user would
                    # be wanting to generate a translated report with holidays
                    # since I believe its main use will be for dates of people

    def write_report(self):
        """
        The short method that runs through each month and creates a page.
        """
        # initialize the dict to fill:
        self.calendar = {}
        # get the information, first from holidays:
        if self.country != 0:
            self.__get_holidays()
        # get data from database:
        self.collect_data()
        # generate the report:
        self.doc.start_paragraph("BIR-Title")
        if self.titletext == _(_TITLE0):
            title = self._("%(str1)s: %(str2)s") % {
                "str1": self._(_TITLE0),
                "str2": self._get_date(Date(self.year)),
            }  # localized year
        else:
            title = self._("%(str1)s: %(str2)s") % {
                "str1": str(self.titletext),
                "str2": self._get_date(Date(self.year)),
            }
        mark = IndexMark(title, INDEX_TYPE_TOC, 1)
        self.doc.write_text(title, mark)
        self.doc.end_paragraph()
        if self.text1.strip() != "":
            self.doc.start_paragraph("BIR-Text1style")
            text1 = str(self.text1)
            if text1 == _(_TITLE1):
                text1 = self._(_TITLE1)
            self.doc.write_text(text1)
            self.doc.end_paragraph()
        if self.text2.strip() != "":
            self.doc.start_paragraph("BIR-Text2style")
            text2 = str(self.text2)
            if text2 == _(_TITLE2):
                text2 = self._(_TITLE2)
            self.doc.write_text(text2)
            self.doc.end_paragraph()
        if self.text3.strip() != "":
            self.doc.start_paragraph("BIR-Text3style")
            self.doc.write_text(str(self.text3))
            self.doc.end_paragraph()
        if self.relationships:
            name = self.center_person.get_primary_name()
            self.doc.start_paragraph("BIR-Text3style")
            mark = utils.get_person_mark(self.database, self.center_person)
            # feature request 2356: avoid genitive form
            self.doc.write_text(
                self._("Relationships shown are to %s")
                % self._name_display.display_name(name),
                mark,
            )
            self.doc.end_paragraph()
        with self._user.progress(
            _("Birthday and Anniversary Report"), _("Formatting months..."), 12
        ) as step:
            for month in range(1, 13):
                step()
                self.print_page(month)

    def print_page(self, month):
        """Prints a month as a page"""
        year = self.year
        dd = self._locale.date_displayer
        self.doc.start_paragraph("BIR-Monthstyle")
        self.doc.write_text(dd.long_months[month].capitalize())
        self.doc.end_paragraph()
        current_date = datetime.date(year, month, 1)
        current_ord = current_date.toordinal()
        started_day = {}
        for i in range(31):
            thisday = current_date.fromordinal(current_ord)
            if thisday.month == month:
                list = self.calendar.get(month, {}).get(thisday.day, [])
                for p, p_obj in list:
                    mark = utils.get_person_mark(self.database, p_obj)
                    p = p.replace("\n", " ")
                    if thisday not in started_day:
                        self.doc.start_paragraph("BIR-Daystyle")
                        self.doc.write_text(str(thisday.day))
                        self.doc.end_paragraph()
                        started_day[thisday] = 1
                    self.doc.start_paragraph("BIR-Datastyle")
                    self.doc.write_text(p, mark)
                    self.doc.end_paragraph()
            current_ord += 1

    def collect_data(self):
        """
        This method runs through the data, and collects the relevant dates
        and text.
        """
        people = self.database.iter_person_handles()
        people = self.filter.apply(self.database, people, user=self._user)

        ngettext = self._locale.translation.ngettext  # to see "nearby" comments
        rel_calc = get_relationship_calculator(reinit=True, clocale=self._locale)

        with self._user.progress(
            _("Birthday and Anniversary Report"), _("Reading database..."), len(people)
        ) as step:
            for person_handle in people:
                step()
                person = self.database.get_person_from_handle(person_handle)
                short_name = self.get_name(person)
                birth_ref = person.get_birth_ref()
                birth_date = None
                if birth_ref:
                    birth_event = self.database.get_event_from_handle(birth_ref.ref)
                    birth_date = birth_event.get_date_object()

                if self.birthdays and birth_date is not None and birth_date.is_valid():
                    birth_date = gregorian(birth_date)

                    year = birth_date.get_year()
                    month = birth_date.get_month()
                    day = birth_date.get_day()

                    prob_alive_date = Date(self.year, month, day)
                    month, day = birth_date.anniversary(self.year)
                    nyears = self.year - year
                    # add some things to handle maiden name:
                    father_lastname = None  # husband, actually
                    if self.maiden_name in [
                        "spouse_first",
                        "spouse_last",
                    ]:  # get husband's last name:
                        if person.get_gender() == Person.FEMALE:
                            family_list = person.get_family_handle_list()
                            if len(family_list) > 0:
                                if self.maiden_name == "spouse_first":
                                    fhandle = family_list[0]
                                else:
                                    fhandle = family_list[-1]
                                fam = self.database.get_family_from_handle(fhandle)
                                father_handle = fam.get_father_handle()
                                mother_handle = fam.get_mother_handle()
                                if mother_handle == person_handle:
                                    if father_handle:
                                        father = self.database.get_person_from_handle(
                                            father_handle
                                        )
                                        if father is not None:
                                            primary_name = father.get_primary_name()
                                            if primary_name:
                                                father_lastname = Surname.get_surname(
                                                    primary_name.get_primary_surname()
                                                )

                    short_name = self.get_name(person, father_lastname)

                    alive = probably_alive(person, self.database, prob_alive_date)
                    if (self.alive and alive) or not self.alive:
                        comment = ""
                        if self.relationships:
                            relation = rel_calc.get_one_relationship(
                                self.database,
                                self.center_person,
                                person,
                                olocale=self._locale,
                            )
                            if relation:
                                # FIXME this won't work for RTL languages
                                comment = " --- %s" % relation
                        deadtxt = ""
                        if not alive:
                            deadtxt = self.deadtxt
                        yeartxt = ""
                        if self.showyear:
                            yeartxt = "(%s) " % year
                        if nyears == 0:
                            text = self._("* %(person)s, birth %(relation)s") % {
                                "person": short_name,
                                "relation": comment,
                            }
                        else:
                            # Translators: leave all/any {...} untranslated
                            text = ngettext(
                                "* {year}{person}{dead}, {age}{relation}",
                                "* {year}{person}{dead}, {age}{relation}",
                                nyears,
                            ).format(
                                year=yeartxt,
                                person=short_name,
                                dead=deadtxt,
                                age=nyears,
                                relation=comment,
                            )

                        self.add_day_item(text, month, day, person)
                if self.anniversaries:
                    family_list = person.get_family_handle_list()
                    for fhandle in family_list:
                        fam = self.database.get_family_from_handle(fhandle)
                        father_handle = fam.get_father_handle()
                        mother_handle = fam.get_mother_handle()
                        if father_handle == person.get_handle():
                            spouse_handle = mother_handle
                        else:
                            continue  # with next person if the father is not "person"
                            # this will keep from duplicating the anniversary
                        if spouse_handle:
                            spouse = self.database.get_person_from_handle(spouse_handle)
                            if spouse:
                                spouse_name = self.get_name(spouse)
                                short_name = self.get_name(person)
                                # TEMP: this will handle ordered events
                                # Gramps 3.0 will have a new mechanism for start/stop events
                                are_married = None
                                for event_ref in fam.get_event_ref_list():
                                    event = self.database.get_event_from_handle(
                                        event_ref.ref
                                    )
                                    if event.type in [
                                        EventType.MARRIAGE,
                                        EventType.MARR_ALT,
                                    ]:
                                        are_married = event
                                    elif event.type in [
                                        EventType.DIVORCE,
                                        EventType.ANNULMENT,
                                        EventType.DIV_FILING,
                                    ]:
                                        are_married = None
                                if are_married is not None:
                                    for event_ref in fam.get_event_ref_list():
                                        event = self.database.get_event_from_handle(
                                            event_ref.ref
                                        )
                                        event_obj = event.get_date_object()
                                        if (
                                            event_obj is not Date.EMPTY
                                            and event_obj.is_valid()
                                        ):
                                            event_obj = gregorian(event_obj)
                                        year = event_obj.get_year()
                                        month = event_obj.get_month()
                                        day = event_obj.get_day()
                                        nyears = self.year - year

                                        if event_obj.is_valid():
                                            prob_alive_date = Date(
                                                self.year, month, day
                                            )
                                            alive1 = probably_alive(
                                                person, self.database, prob_alive_date
                                            )
                                            alive2 = probably_alive(
                                                spouse, self.database, prob_alive_date
                                            )
                                            deadtxt1 = ""
                                            deadtxt2 = ""
                                            if not alive1:
                                                deadtxt1 = self.deadtxt
                                            if not alive2:
                                                deadtxt2 = self.deadtxt
                                            yeartxt = ""
                                            if self.showyear:
                                                yeartxt = "(%s) " % year
                                            if nyears == 0:
                                                text = self._(
                                                    "⚭ %(spouse)s and\n %(person)s, wedding"
                                                ) % {
                                                    "spouse": spouse_name,
                                                    "person": short_name,
                                                }
                                            else:
                                                # Translators: leave all/any {...} untranslated
                                                text = ngettext(
                                                    "⚭ {year}{spouse}{deadtxt2} and\n {person}{deadtxt1}, {nyears}",
                                                    "⚭ {year}{spouse}{deadtxt2} and\n {person}{deadtxt1}, {nyears}",
                                                    nyears,
                                                ).format(
                                                    year=yeartxt,
                                                    spouse=spouse_name,
                                                    deadtxt2=deadtxt2,
                                                    person=short_name,
                                                    deadtxt1=deadtxt1,
                                                    nyears=nyears,
                                                )
                                                if (
                                                    self.alive and alive1 and alive2
                                                ) or not self.alive:
                                                    self.add_day_item(
                                                        text, month, day, spouse
                                                    )

                death_ref = person.get_death_ref()
                death_date = None
                if death_ref:
                    death_event = self.database.get_event_from_handle(death_ref.ref)
                    death_date = death_event.get_date_object()

                if (
                    self.death_anniversaries
                    and death_date is not None
                    and death_date.is_valid()
                ):
                    death_date = gregorian(death_date)

                    year = death_date.get_year()
                    month = death_date.get_month()
                    day = death_date.get_day()

                    nyears = self.year - year

                    comment = ""
                    if self.relationships:
                        relation = rel_calc.get_one_relationship(
                            self.database,
                            self.center_person,
                            person,
                            olocale=self._locale,
                        )
                        if relation:
                            # FIXME this won't work for RTL languages
                            comment = " --- %s" % relation
                    yeartxt = ""
                    if self.showyear:
                        yeartxt = "(%s) " % year
                    if nyears == 0:
                        text = _("✝ {person}, death {relation}").format(
                            person=short_name, relation=comment
                        )
                    else:
                        text = ngettext(
                            "✝ {year}{person}, {age}{relation}",
                            "✝ {year}{person}, {age}{relation}",
                            nyears,
                        ).format(
                            year=yeartxt,
                            person=short_name,
                            age=nyears,
                            relation=comment,
                        )
                    self.add_day_item(text, month, day, person)


# ------------------------------------------------------------------------
#
# BirthdayOptions
#
# ------------------------------------------------------------------------
class BirthdayOptions(MenuReportOptions):
    """Options for the Birthday/Anniversary Report"""

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__pid = None
        self.__filter = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        return self.__filter.get_filter().get_name()

    def add_menu_options(self, menu):
        """Add the options for the text birthday report"""
        category_name = _("Report Options")

        self.__filter = FilterOption(_("Filter"), 0)
        self.__filter.set_help(_("Select the filter to be applied to the report."))
        menu.add_option(category_name, "filter", self.__filter)
        self.__filter.connect("value-changed", self.__filter_changed)

        self.__pid = PersonOption(_("Filter Person"))
        self.__pid.set_help(_("The center person for the filter."))
        menu.add_option(category_name, "pid", self.__pid)
        self.__pid.connect("value-changed", self.__update_filters)

        titletext = StringOption(_("Title text"), _(_TITLE0))
        titletext.set_help(_("Title of report"))
        menu.add_option(category_name, "titletext", titletext)

        text1 = StringOption(_("Text Area 1"), _(_TITLE1))
        text1.set_help(_("First line of text at bottom of report"))
        menu.add_option(category_name, "text1", text1)

        text2 = StringOption(_("Text Area 2"), _(_TITLE2))
        text2.set_help(_("Second line of text at bottom of report"))
        menu.add_option(category_name, "text2", text2)

        text3 = StringOption(
            _("Text Area 3"),
            URL_HOMEPAGE,
        )
        text3.set_help(_("Third line of text at bottom of report"))
        menu.add_option(category_name, "text3", text3)

        category_name = _("Report Options (2)")

        self._nf = stdoptions.add_name_format_option(menu, category_name)
        self._nf.connect("value-changed", self.__update_filters)

        stdoptions.add_private_data_option(menu, category_name)

        alive = BooleanOption(_("Include only living people"), True)
        alive.set_help(_("Include only living people in the report"))
        menu.add_option(category_name, "alive", alive)

        stdoptions.add_gramps_id_option(menu, category_name)

        deadtxt = StringOption(_("Dead Symbol"), _(_DEADTXT))
        deadtxt.set_help(_("This will show after name to indicate that person is dead"))
        menu.add_option(category_name, "deadtxt", deadtxt)

        self.__update_filters()

        stdoptions.add_localization_option(menu, category_name)

        showyear = BooleanOption(_("Show event year"), True)
        showyear.set_help(_("Prints the year the event took place in the report"))
        menu.add_option(category_name, "showyear", showyear)

        category_name = _("Content")

        year = NumberOption(_("Year of report"), time.localtime()[0], 1000, 3000)
        year.set_help(_("Year of report"))
        menu.add_option(category_name, "year", year)

        country = EnumeratedListOption(_("Country for holidays"), 0)
        holiday_table = libholiday.HolidayTable()
        countries = holiday_table.get_countries()
        countries.sort()
        if len(countries) == 0 or (len(countries) > 0 and countries[0] != ""):
            countries.insert(0, "")
        count = 0
        for c in countries:
            country.add_item(count, c)
            count += 1
        country.set_help(_("Select the country to see associated holidays"))
        menu.add_option(category_name, "country", country)

        maiden_name = EnumeratedListOption(_("Birthday surname"), "own")
        maiden_name.add_item(
            "spouse_first", _("Wives use husband's surname (from first family listed)")
        )
        maiden_name.add_item(
            "spouse_last", _("Wives use husband's surname (from last family listed)")
        )
        maiden_name.add_item("own", _("Wives use their own surname"))
        maiden_name.set_help(_("Select married women's displayed surname"))
        menu.add_option(category_name, "maiden_name", maiden_name)

        birthdays = BooleanOption(_("Include birthdays"), True)
        birthdays.set_help(_("Whether to include birthdays"))
        menu.add_option(category_name, "birthdays", birthdays)

        anniversaries = BooleanOption(_("Include anniversaries"), True)
        anniversaries.set_help(_("Whether to include anniversaries"))
        menu.add_option(category_name, "anniversaries", anniversaries)

        death_anniversaries = BooleanOption(_("Include death anniversaries"), True)
        death_anniversaries.set_help(_("Whether to include anniversaries of death"))
        menu.add_option(category_name, "death_anniversaries", death_anniversaries)

        show_relships = BooleanOption(_("Include relationship to center person"), False)
        show_relships.set_help(
            _("Whether to include relationships to the center person")
        )
        menu.add_option(category_name, "relationships", show_relships)

    def __update_filters(self):
        """
        Update the filter list based on the selected person
        """
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        nfv = self._nf.get_value()
        filter_list = utils.get_person_filters(
            person, include_single=False, name_format=nfv
        )
        self.__filter.set_filters(filter_list)

    def __filter_changed(self):
        """
        Handle filter change. If the filter is not specific to a person,
        disable the person option
        """
        filter_value = self.__filter.get_value()
        if filter_value == 0:  # "Entire Database" (as "include_single=False")
            self.__pid.set_available(False)
        else:
            # The other filters need a center person (assume custom ones too)
            self.__pid.set_available(True)

    def make_my_style(
        self,
        default_style,
        name,
        description,
        size=9,
        font=FONT_SERIF,
        justified="left",
        color=None,
        align=PARA_ALIGN_CENTER,
        shadow=None,
        italic=0,
        bold=0,
        borders=0,
        indent=None,
    ):
        """Create paragraph and graphic styles of the same name"""
        # Paragraph:
        f = FontStyle()
        f.set_size(size)
        f.set_type_face(font)
        f.set_italic(italic)
        f.set_bold(bold)
        p = ParagraphStyle()
        p.set_font(f)
        p.set_alignment(align)
        p.set_description(description)
        p.set_top_border(borders)
        p.set_left_border(borders)
        p.set_bottom_border(borders)
        p.set_right_border(borders)
        if indent:
            p.set(first_indent=indent)
        if justified == "left":
            p.set_alignment(PARA_ALIGN_LEFT)
        elif justified == "right":
            p.set_alignment(PARA_ALIGN_RIGHT)
        elif justified == "center":
            p.set_alignment(PARA_ALIGN_CENTER)
        default_style.add_paragraph_style(name, p)
        # Graphics:
        g = GraphicsStyle()
        g.set_paragraph_style(name)
        if shadow:
            g.set_shadow(*shadow)
        if color is not None:
            g.set_fill_color(color)
        if not borders:
            g.set_line_width(0)
        default_style.add_draw_style(name, g)

    def make_default_style(self, default_style):
        """Add the styles used in this report"""
        self.make_my_style(
            default_style,
            "BIR-Title",
            _("Title text style"),
            14,
            bold=1,
            justified="center",
        )
        self.make_my_style(
            default_style, "BIR-Datastyle", _("Data text display"), 12, indent=1.0
        )
        self.make_my_style(
            default_style,
            "BIR-Daystyle",
            _("Day text style"),
            12,
            indent=0.5,
            italic=1,
            bold=1,
        )
        self.make_my_style(
            default_style, "BIR-Monthstyle", _("Month text style"), 14, bold=1
        )
        self.make_my_style(
            default_style,
            "BIR-Text1style",
            _("Text at bottom, line 1"),
            12,
            justified="center",
        )
        self.make_my_style(
            default_style,
            "BIR-Text2style",
            _("Text at bottom, line 2"),
            12,
            justified="center",
        )
        self.make_my_style(
            default_style,
            "BIR-Text3style",
            _("Text at bottom, line 3"),
            12,
            justified="center",
        )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Utility functions to cast types
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import os
import sys
import logging

LOG = logging.getLogger(".")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


def cast_to_bool(val):
    return val in ["True", "true", _("True"), _("true"), "1"]  # 3139


def get_type_converter(val):
    """
    Return function that converts strings into the type of val.
    """
    val_type = type(val)
    if isinstance(val, str):
        return str
    elif val_type == int:
        return int
    elif val_type == float:
        return float
    elif val_type == bool:
        return cast_to_bool
    elif val_type in (list, tuple):
        return list


def type_name(val):
    """
    Return the name the type of val.

    Only numbers and strings are supported.
    The rest becomes strings (unicode).
    """
    val_type = type(val)
    if val_type == int:
        return "int"
    elif val_type == float:
        return "float"
    elif val_type == bool:
        return "bool"
    elif isinstance(val, str):
        return "unicode"
    return "unicode"


def get_type_converter_by_name(val_str):
    """
    Return function that converts strings into the type given by val_str.

    Only numbers and strings are supported.
    The rest becomes strings (unicode).
    """
    if val_str == "int":
        return int
    elif val_str == "float":
        return float
    elif val_str == "bool":
        return cast_to_bool
    elif val_str in ("str", "unicode"):
        return str
    return str



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2007  Donald N. Allingham
# Copyright (C) 2009-2010  Nick Hall
# Copyright (C) 2009       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide the base classes for GRAMPS' DataView classes
"""

# ----------------------------------------------------------------
#
# python modules
#
# ----------------------------------------------------------------
from __future__ import annotations
from abc import abstractmethod
import os
import pickle
from time import perf_counter
import logging
from collections import deque

LOG = logging.getLogger(".gui.listview")

# ----------------------------------------------------------------
#
# gtk
#
# ----------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

# ----------------------------------------------------------------
#
# Gramps
#
# ----------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from .pageview import PageView
from .navigationview import NavigationView
from ..uimanager import ActionGroup
from ..columnorder import ColumnOrder
from ..csvdialect import CsvDialect
from gramps.gen.config import config
from gramps.gen.db import DbTxn
from gramps.gen.errors import WindowActiveError, FilterError, HandleError
from ..filters import SearchBar
from ..widgets.menuitem import add_menuitem
from gramps.gen.const import CUSTOM_FILTERS
from gramps.gen.utils.debug import profile
from gramps.gen.utils.string import data_recover_msg
from gramps.gen.plug import CATEGORY_QR_MISC
from ..dialog import QuestionDialog, QuestionDialog3, ErrorDialog, MultiSelectDialog
from ..editors import FilterEditor
from ..ddtargets import DdTargets
from ..plug.quick import create_quickreport_menu, create_web_connect_menu
from ..utils import is_right_click
from ..widgets.interactivesearchbox import InteractiveSearchBox
from ..widgets.persistenttreeview import PersistentTreeView

# ----------------------------------------------------------------
#
# Constants
#
# ----------------------------------------------------------------
TEXT = 1
MARKUP = 2
ICON = 3


# ----------------------------------------------------------------
#
# ListView
#
# ----------------------------------------------------------------
class ListView(NavigationView):
    COLUMNS: list[tuple[str, int, str | None]] = []
    # listview config settings that are always present related to the columns
    CONFIGSETTINGS = (
        ("columns.visible", []),
        ("columns.rank", []),
        ("columns.size", []),
    )
    ADD_MSG = ""
    EDIT_MSG = ""
    DEL_MSG = ""
    MERGE_MSG = ""
    FILTER_TYPE = ""  # Set in inheriting class
    QR_CATEGORY = CATEGORY_QR_MISC

    def __init__(
        self,
        title,
        pdata,
        dbstate,
        uistate,
        make_model,
        signal_map,
        bm_type,
        nav_group,
        multiple=False,
        filter_class=None,
    ):
        NavigationView.__init__(
            self, title, pdata, dbstate, uistate, bm_type, nav_group
        )
        # default is listviews keep themself in sync with database
        self._dirty_on_change_inactive = False

        self.filter_class = filter_class
        self.pb_renderer = Gtk.CellRendererPixbuf()
        self.renderer = Gtk.CellRendererText()
        self.renderer.set_property("ellipsize", Pango.EllipsizeMode.END)
        self.sort_col = 0
        self.sort_order = Gtk.SortType.ASCENDING
        self.columns = []
        self.make_model = make_model
        self.model = None
        self.signal_map = signal_map
        self.multiple_selection = multiple
        self.generic_filter = None
        dbstate.connect("database-changed", self.change_db)
        self.connect_signals()
        self.at_popup_action = None
        self.at_popup_menu = None

    def no_database(self):
        ## TODO GTK3: This is never called!! Dbguielement disconnects
        ## signals on database changed, so it cannot be called
        ## Undo part of Revision 20296 if all works good.
        self.list.set_model(None)
        self.model.destroy()
        self.model = None
        self.build_tree()

    ####################################################################
    # Build interface
    ####################################################################
    def build_widget(self):
        """
        Builds the interface and returns a Gtk.Container type that
        contains the interface. This containter will be inserted into
        a Gtk.Notebook page.
        """
        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.vbox.set_border_width(4)
        self.vbox.set_spacing(4)

        self.search_bar = SearchBar(self.dbstate, self.uistate, self.search_build_tree)
        filter_box = self.search_bar.build()

        self.list = PersistentTreeView(self.uistate, self.get_config_name())
        self.list.set_headers_visible(True)
        self.list.set_headers_clickable(True)
        self.list.set_fixed_height_mode(True)
        self.list.connect("button-press-event", self._button_press)
        self.list.connect("key-press-event", self._key_press)
        self.list.connect("start-interactive-search", self.open_all_nodes)
        self.searchbox = InteractiveSearchBox(self.list)

        if self.drag_info():
            self.list.connect("drag_data_get", self.drag_data_get)
            self.list.connect("drag_begin", self.drag_begin)
        if self.drag_dest_info():
            self.list.connect("drag_data_received", self.drag_data_received)
            self.list.drag_dest_set(
                Gtk.DestDefaults.MOTION | Gtk.DestDefaults.DROP,
                [self.drag_dest_info().target()],
                Gdk.DragAction.MOVE | Gdk.DragAction.COPY,
            )

        scrollwindow = Gtk.ScrolledWindow()
        scrollwindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrollwindow.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        scrollwindow.add(self.list)

        self.vbox.pack_start(filter_box, False, True, 0)
        self.vbox.pack_start(scrollwindow, True, True, 0)

        self.renderer = Gtk.CellRendererText()
        self.renderer.set_property("ellipsize", Pango.EllipsizeMode.END)

        self.columns = []
        self.build_columns()
        self.selection = self.list.get_selection()
        if self.multiple_selection:
            self.selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        self.selection.connect("changed", self.row_changed)

        self.setup_searches()
        self.list.restore_column_size()
        return self.vbox

    def define_actions(self):
        """
        Required define_actions function for PageView. Builds the action
        group information required. We extend beyond the normal here,
        since we want to have more than one action group for the PersonView.
        Most PageViews really won't care about this.
        """

        NavigationView.define_actions(self)

        self.edit_action = ActionGroup(name=self.title + "/Edits")
        self.edit_action.add_actions(
            [
                ("Add", self.add, "<Primary>Insert"),
                ("Remove", self.remove, "<Primary>Delete"),
                ("PRIMARY-BackSpace", self.remove, "<PRIMARY>BackSpace"),
                ("Merge", self.merge),
            ]
        )

        self._add_action_group(self.edit_action)
        self.action_list.extend(
            [
                ("ExportTab", self.export),
                ("Edit", self.edit, "<Primary>Return"),
                ("PRIMARY-J", self.jump, "<PRIMARY>J"),
                ("FilterEdit", self.filter_editor),
            ]
        )

    def build_columns(self, preserve_col=True):
        """
        build the columns
        """
        # Preserve the column widths if rebuilding the view.
        # removed since we have a PersistentTreeView
        # if self.columns and preserve_col:
        #    self.list.save_column_info(self.list)
        list(map(self.list.remove_column, self.columns))

        self.columns = []

        index = 0
        for pair in self.column_order():
            if not pair[0]:
                continue
            col_name, col_type, col_icon = self.COLUMNS[pair[1]]

            if col_type == ICON:
                column = Gtk.TreeViewColumn(col_name, self.pb_renderer)
                column.set_cell_data_func(self.pb_renderer, self.icon, pair[1])
            else:
                column = Gtk.TreeViewColumn(col_name, self.renderer)
                if col_type == MARKUP:
                    column.add_attribute(self.renderer, "markup", pair[1])
                else:
                    column.add_attribute(self.renderer, "text", pair[1])

            if col_icon is not None:
                image = Gtk.Image()
                image.set_from_icon_name(col_icon, Gtk.IconSize.MENU)
                image.set_tooltip_text(col_name)
                image.show()
                column.set_widget(image)

            if self.model and self.model.color_column() is not None:
                column.set_cell_data_func(self.renderer, self.foreground_color)

            column.connect("clicked", self.column_clicked, index)

            column.set_resizable(True)
            column.set_clickable(True)
            column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
            column.set_fixed_width(pair[2])

            self.columns.append(column)
            self.list.append_column(column)
            index += 1
        return self.vbox

    def icon(self, column, renderer, model, iter_, col_num):
        """
        Set the icon-name property of the cell renderer.  We use a cell data
        function because there is a problem returning None from a model.
        """
        icon_name = model.get_value(iter_, col_num)
        if icon_name == "":
            icon_name = None
        renderer.set_property("icon-name", icon_name)

    def foreground_color(self, column, renderer, model, iter_, data=None):
        """
        Set the foreground color of the cell renderer.  We use a cell data
        function because there is a problem returning None from a model.
        """
        fg_color = model.get_value(iter_, model.color_column())
        if fg_color == "":
            fg_color = None
        renderer.set_property("foreground", fg_color)

    def set_active(self):
        """
        Called when the page is displayed.
        """
        NavigationView.set_active(self)
        self.uistate.viewmanager.tags.tag_enable(update_menu=False)
        self.uistate.show_filter_results(
            self.dbstate, self.model.displayed(), self.model.total()
        )

    def set_inactive(self):
        """
        Called when the page is no longer displayed.
        """
        NavigationView.set_inactive(self)
        self.uistate.viewmanager.tags.tag_disable()

    def build_tree(self, force_sidebar=False, preserve_col=True):
        if self.list.get_columns()[0].get_width() > 0:
            self.list.save_column_info(self.list)
        if self.active:
            cput0 = perf_counter()
            if not self.search_bar.is_visible():
                filter_info = (True, self.generic_filter, False)
            else:
                filter_info = self.search_bar.get_value()
                if not filter_info[0]:
                    filter_info = (
                        filter_info[0],
                        filter_info[1],
                        filter_info[1][0] in self.exact_search(),
                    )

            if self.dirty or not self.model:
                if self.model:
                    self.list.set_model(None)
                    self.model.destroy()
                self.model = self.make_model(
                    self.dbstate.db,
                    self.uistate,
                    self.sort_col,
                    search=filter_info,
                    sort_map=self.column_order(),
                )
            else:
                # the entire data to show is already in memory.
                # run only the part that determines what to show
                self.list.set_model(None)
                self.model.set_search(filter_info)
                try:
                    self.model.rebuild_data()
                except FilterError as msg:
                    (msg1, msg2) = msg.messages()
                    ErrorDialog(msg1, msg2, parent=self.uistate.window)

            cput1 = perf_counter()
            self.build_columns(preserve_col)
            self.list.restore_column_size()
            cput2 = perf_counter()
            self.list.set_model(self.model)
            cput3 = perf_counter()
            self.__display_column_sort()
            self.goto_active(None)

            self.dirty = False
            cput4 = perf_counter()
            self.uistate.show_filter_results(
                self.dbstate, self.model.displayed(), self.model.total()
            )
            LOG.debug(
                self.__class__.__name__
                + " build_tree "
                + str(perf_counter() - cput0)
                + " sec"
            )
            LOG.debug(
                "parts "
                + str(cput1 - cput0)
                + " , "
                + str(cput2 - cput1)
                + " , "
                + str(cput3 - cput2)
                + " , "
                + str(cput4 - cput3)
                + " , "
                + str(perf_counter() - cput4)
            )

        else:
            self.dirty = True

    def search_build_tree(self):
        self.build_tree()

    def exact_search(self):
        """
        Returns a tuple indicating columns requiring an exact search
        """
        return ()

    def get_viewtype_stock(self):
        """Type of view in category, default listview is a flat list"""
        return "gramps-tree-list"

    def filter_editor(self, *obj):
        try:
            FilterEditor(self.FILTER_TYPE, CUSTOM_FILTERS, self.dbstate, self.uistate)
        except WindowActiveError:
            return

    def setup_searches(self):
        """Build the default filters and add them to the filter menu."""
        self.search_bar.setup_searches(
            [
                (self.COLUMNS[pair[1]][0], pair[1], pair[1] in self.exact_search())
                for pair in self.column_order()
                if pair[0]
            ]
        )

    def sidebar_toggled(self, active, data=None):
        """
        Called when the sidebar is toggled.
        """
        if active:
            self.search_bar.hide()
        else:
            self.search_bar.show()

    ####################################################################
    # Navigation
    ####################################################################
    def goto_handle(self, handle):
        """
        Go to a given handle in the list.
        Required by the NavigationView interface.

        We have a bit of a problem due to the nature of how GTK works.
        We have to unselect the previous path and select the new path. However,
        these cause a row change, which calls the row_change callback, which
        can end up calling change_active_person, which can call
        goto_active_person, causing a bit of recusion. Confusing, huh?

        Unfortunately, row_change has to be able to call change_active_person,
        because this can occur from the interface in addition to programatically.

        To handle this, we set the self.inactive variable that we can check
        in row_change to look for this particular condition.
        """
        if not handle or handle in self.selected_handles():
            return

        iter_ = self.model.get_iter_from_handle(handle)
        if iter_:
            if not (self.model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY):
                # Expand tree
                parent_iter = self.model.iter_parent(iter_)
                if parent_iter:
                    parent_path = self.model.get_path(parent_iter)
                    if parent_path:
                        parent_path_list = parent_path.get_indices()
                        for i, value in enumerate(parent_path_list):
                            expand_path = Gtk.TreePath(
                                tuple([x for x in parent_path_list[: i + 1]])
                            )
                            self.list.expand_row(expand_path, False)

            # Select active object
            path = self.model.get_path(iter_)
            self.selection.unselect_all()
            self.selection.select_path(path)
            self.list.scroll_to_cell(path, None, 1, 0.5, 0)
        else:
            self.selection.unselect_all()
            self.uistate.push_message(self.dbstate, _("Active object not visible"))

    def add_bookmark(self, *obj):
        mlist = []
        self.selection.selected_foreach(self.blist, mlist)

        if mlist:
            self.bookmarks.add(mlist[0])
        else:
            from ..dialog import WarningDialog

            WarningDialog(
                _("Could Not Set a Bookmark"),
                _("A bookmark could not be set because " "nothing was selected."),
                parent=self.uistate.window,
            )

    ####################################################################
    #
    ####################################################################

    def drag_info(self):
        """
        Specify the drag type for a single selected row
        """
        return None

    def drag_list_info(self):
        """
        Specify the drag type for a multiple selected rows
        """
        return DdTargets.LINK_LIST

    def drag_dest_info(self):
        """
        Specify the drag type for objects dropped on the view
        """
        return None

    def drag_begin(self, widget, context):
        widget.drag_source_set_icon_name(self.get_stock())

    def drag_data_get(self, widget, context, sel_data, info, time):
        selected_ids = self.selected_handles()

        # Gtk.selection_add_target(widget, sel_data.get_selection(),
        #                         Gdk.atom_intern(self.drag_info().drag_type, False),
        #                         self.drag_info().app_id)

        if len(selected_ids) == 1:
            data = (self.drag_info().drag_type, id(self), selected_ids[0], 0)
            sel_data.set(self.drag_info().atom_drag_type, 8, pickle.dumps(data))
        elif len(selected_ids) > 1:
            data = (
                self.drag_list_info().drag_type,
                id(self),
                [(self.drag_list_info().drag_type, handle) for handle in selected_ids],
                0,
            )
            sel_data.set(self.drag_list_info().atom_drag_type, 8, pickle.dumps(data))
        else:
            # pass empty
            data = (self.drag_info().drag_type, id(self), [], 0)
            sel_data.set(self.drag_list_info().atom_drag_type, 8, pickle.dumps(data))

    def set_column_order(self):
        """
        change the order of the columns to that given in config file
        after config file changed. We reset the sort to the first column
        """
        # now we need to rebuild the model so it contains correct column info
        self.dirty = True
        # make sure we sort on first column. We have no idea where the
        # column that was sorted on before is situated now.
        self.sort_col = 0
        self.sort_order = Gtk.SortType.ASCENDING
        self.setup_searches()
        self.build_tree(preserve_col=False)

    def column_order(self):
        """
        Column order is obtained from the config file of the listview.
        A column order is a list of 3-tuples. The order in the list is the
        order the columns must appear in.
        For a column, the 3-tuple should be (enable, modelcol, sizecol), where
            enable: show this column or don't show it
            modelcol: column in the datamodel this column is build of
            size: size the column should have
        """
        order = self._config.get("columns.rank")
        size = self._config.get("columns.size")
        vis = self._config.get("columns.visible")

        colord = [(1 if val in vis else 0, val, size) for val, size in zip(order, size)]
        return colord

    def get_column_widths(self):
        return [column.get_width() for column in self.columns]

    ####################################################################
    # Object Delete functions
    ####################################################################
    # def remove(self):  # this is over-ridden in each view
    # """
    # must return the tuple of (object type and handle) for each
    # selected item
    # """
    # handles = self.selected_handles()
    # ht_list = [('Person', hndl) for hndl in handles]
    # self.remove_selected_objects(ht_list)

    def remove_selected_objects(self, ht_list=None):
        """
        Function to remove selected objects
        """
        if len(ht_list) == 1:
            obj = self.dbstate.db.method("get_%s_from_handle", ht_list[0][0])(
                ht_list[0][1]
            )
            msg1 = self._message1_format(obj)
            msg2 = self._message2_format(obj)
            msg2 = "%s %s" % (msg2, data_recover_msg)
            QuestionDialog(
                msg1,
                msg2,
                _("_Delete"),
                lambda: self.delete_object_response(obj, parent=self.uistate.window),
                parent=self.uistate.window,
            )
        else:
            MultiSelectDialog(
                self._message1_format,
                self._message2_format,
                ht_list,
                lambda x: self.dbstate.db.method("get_%s_from_handle", x[0])(x[1]),
                yes_func=self.delete_object_response,
                multi_yes_func=self.delete_multi_object_response,
                parent=self.uistate.window,
            )

    def _message1_format(self, obj):
        """
        Header format for remove dialogs.
        """
        return _("Delete {type} [{gid}]?").format(
            type=_(obj.__class__.__name__), gid=obj.gramps_id
        )

    def _message2_format(self, _obj):
        """
        Detailed message format for the remove dialogs.
        """
        return _("Deleting item will remove it from the database.")

    def _message3_format(self, obj):
        """
        Transaction label format
        """
        return "%s %s [%s]" % (
            _("Delete"),
            _(obj.__class__.__name__),
            obj.get_gramps_id(),
        )

    def delete_object_response(self, obj, parent=None):
        """
        Delete the object from the database.
        """
        with DbTxn(self._message3_format(obj), self.dbstate.db) as trans:
            # self.db.disable_signals()
            self.remove_object_from_handle(
                obj.__class__.__name__,
                obj.handle,
                trans,
                in_use_prompt=True,
                parent=parent,
            )
            # self.dbstate.db.enable_signals()
        self.uistate.set_busy_cursor(False)

    def delete_multi_object_response(self, ht_list=None, parent=None):
        """
        Deletes multiple objects from the database.
        """
        # set the busy cursor, so the user knows that we are working
        self.uistate.set_busy_cursor(True)
        self.uistate.progress.show()
        self.uistate.push_message(self.dbstate, _("Processing..."))
        hndl_cnt = len(ht_list) / 100
        _db = self.dbstate.db
        _db.disable_signals()

        # create the transaction
        with DbTxn("", _db) as trans:
            for indx, item in enumerate(ht_list):
                result = self.remove_object_from_handle(
                    *item, trans, in_use_prompt=False, parent=parent
                )
                self.uistate.pulse_progressbar(indx / hndl_cnt)
                if result == -1:
                    break
            trans.set_description(_("Multiple Selection Delete"))

        _db.enable_signals()
        _db.request_rebuild()
        self.uistate.progress.hide()
        self.uistate.set_busy_cursor(False)

    def remove_object_from_handle(
        self, obj_type, handle, trans, in_use_prompt=False, parent=None
    ):
        """
        deletes a single object from database
        """
        obj = self.dbstate.db.method("get_%s_from_handle", obj_type)(handle)
        bl_list = list(self.dbstate.db.find_backlink_handles(handle))
        if in_use_prompt:
            res = self._in_use_prompt(obj, bl_list, parent=parent)
            if res != 1:  # Cancel or No
                return res
        # perfom the cleanup
        for ref_type, ref_hndl in bl_list:
            ref_obj = self.dbstate.db.method("get_%s_from_handle", ref_type)(ref_hndl)
            ref_obj.remove_handle_references(obj_type, [handle])
            self.dbstate.db.method("commit_%s", ref_type)(ref_obj, trans)
        self.dbstate.db.method("remove_%s", obj_type)(obj.get_handle(), trans)

    def _in_use_prompt(self, obj, bl_list, parent=None):
        """
        Prompt user if he wants to continue because in use
        """
        if bl_list:
            msg = _(
                "This item is currently being used. "
                "Deleting it will remove it from the database and "
                "from all other items that reference it."
            )
        else:
            msg = _("Deleting item will remove it from the database.")

        msg += " " + data_recover_msg
        descr = obj.get_gramps_id()
        ques = QuestionDialog3(
            _("Delete %s?") % descr, msg, _("_Yes"), _("_No"), parent=parent
        )
        return ques.run()

    def blist(self, store, path, iter_, sel_list):
        """GtkTreeSelectionForeachFunc
        construct a list sel_list with all selected handles
        """
        handle = store.get_handle_from_iter(iter_)
        if handle is not None:
            sel_list.append(handle)

    def selected_handles(self):
        mlist = []
        if self.selection:
            self.selection.selected_foreach(self.blist, mlist)
        return mlist

    def first_selected(self):
        mlist = []
        self.selection.selected_foreach(self.blist, mlist)
        if mlist:
            return mlist[0]
        else:
            return None

    ####################################################################
    # Signal handlers
    ####################################################################
    def column_clicked(self, obj, data):
        """
        Called when a column is clicked.

        obj     A TreeViewColumn object of the column clicked
        data    The column index
        """
        self.uistate.set_busy_cursor(True)
        self.uistate.push_message(self.dbstate, _("Column clicked, sorting..."))
        cput = perf_counter()
        same_col = False
        if self.sort_col != data:
            order = Gtk.SortType.ASCENDING
        else:
            same_col = True
            if (
                self.columns[data].get_sort_order() == Gtk.SortType.DESCENDING
                or not self.columns[data].get_sort_indicator()
            ):
                order = Gtk.SortType.ASCENDING
            else:
                order = Gtk.SortType.DESCENDING

        self.sort_col = data
        self.sort_order = order
        handle = self.first_selected()

        if not self.search_bar.is_visible():
            filter_info = (True, self.generic_filter, False)
        else:
            value = self.search_bar.get_value()
            filter_info = (False, value, value[0] in self.exact_search())

        if same_col:
            # activate when https://bugzilla.gnome.org/show_bug.cgi?id=684558
            # is resolved
            if False:
                self.model.reverse_order()
            else:
                ## GTK 3.6 rows_reordered not exposed by gi, we need to reconnect
                ## model to obtain desired effect, but this collapses nodes ...
                self.list.set_model(None)
                self.model.reverse_order()
                self.list.set_model(self.model)
        else:
            self.model = self.make_model(
                self.dbstate.db,
                self.uistate,
                self.sort_col,
                self.sort_order,
                search=filter_info,
                sort_map=self.column_order(),
            )

            self.list.set_model(self.model)

        self.__display_column_sort()

        if handle:
            self.goto_handle(handle)

        # set the search column to be the sorted column
        search_col = self.column_order()[data][1]
        self.list.set_search_column(search_col)

        self.uistate.set_busy_cursor(False)

        LOG.debug(
            "   "
            + self.__class__.__name__
            + " column_clicked "
            + str(perf_counter() - cput)
            + " sec"
        )

    def __display_column_sort(self):
        for i, c in enumerate(self.columns):
            c.set_sort_indicator(i == self.sort_col)
        self.columns[self.sort_col].set_sort_order(self.sort_order)

    def connect_signals(self):
        """
        Connect database signals defined in the signal map.
        """
        for sig in self.signal_map:
            self.callman.add_db_signal(sig, self.signal_map[sig])
        self.callman.add_db_signal("tag-update", self.tag_updated)

    def change_db(self, db):
        """
        Called when the database is changed.
        """
        self.list.set_model(None)
        self._change_db(db)
        self.connect_signals()

        if self.active:
            # force rebuild of the model on build of tree
            self.dirty = True
            self.build_tree()
            self.bookmarks.redraw()
        else:
            self.dirty = True

    def row_changed(self, selection):
        """
        Called with a list selection is changed.

        Check the selected objects in the list and return those that have
        handles attached.  Set the active object to the first item in the
        list. If no row is selected, set the active object to None.
        """
        selected_ids = self.selected_handles()
        if len(selected_ids) > 0:
            # In certain cases the tree models do row updates which result in a
            # selection changed signal to a handle in progress of being
            # deleted.  In these cases we don't want to change the active to
            # non-existant handles.
            if hasattr(self.model, "dont_change_active"):
                if not self.model.dont_change_active:
                    self.change_active(selected_ids[0])
            else:
                self.change_active(selected_ids[0])

        if len(selected_ids) == 1:
            if self.drag_info():
                self.list.drag_source_set(
                    Gdk.ModifierType.BUTTON1_MASK,
                    [self.drag_info().target()],
                    Gdk.DragAction.COPY,
                )
        elif len(selected_ids) > 1:
            if self.drag_list_info():
                self.list.drag_source_set(
                    Gdk.ModifierType.BUTTON1_MASK,
                    [self.drag_list_info().target()],
                    Gdk.DragAction.COPY,
                )

        if self.uistate.viewmanager.active_page == self:
            self.uistate.modify_statusbar(self.dbstate)

    def row_add(self, handle_list):
        """
        Called when an object is added.
        """
        if self.active or (not self.dirty and not self._dirty_on_change_inactive):
            cput = perf_counter()
            list(map(self.model.add_row_by_handle, handle_list))
            LOG.debug(
                "   "
                + self.__class__.__name__
                + " row_add "
                + str(perf_counter() - cput)
                + " sec"
            )
            if self.active:
                self.uistate.show_filter_results(
                    self.dbstate, self.model.displayed(), self.model.total()
                )
        else:
            self.dirty = True

    def row_update(self, handle_list):
        """
        Called when an object is updated.
        """
        if self.model:
            self.model.prev_handle = None
        if self.active or (not self.dirty and not self._dirty_on_change_inactive):
            cput = perf_counter()
            # store selected handles
            self._sel_handles_before_update = self.selected_handles()
            list(map(self.model.update_row_by_handle, handle_list))
            LOG.debug(
                "   "
                + self.__class__.__name__
                + " row_update "
                + str(perf_counter() - cput)
                + " sec"
            )
            # Ensure row is still selected after a change of postion in tree.
            if self._sel_handles_before_update:
                # we can only set one selected again, we take last
                self.goto_handle(self._sel_handles_before_update[-1])
            elif handle_list and not self.selected_handles():
                self.goto_handle(handle_list[-1])
        else:
            self.dirty = True

    def row_delete(self, handle_list):
        """
        Called when an object is deleted.
        """
        if self.active or (not self.dirty and not self._dirty_on_change_inactive):
            cput = perf_counter()
            for hndl in handle_list:
                if hndl != handle_list[-1]:
                    # For multiple deletes, row updates can result in a
                    # selection changed signal to a handle already
                    # deleted.  In these cases we don't want to change the
                    # active to non-existant handles.
                    self.model.dont_change_active = True
                else:
                    # Allow active changed on last item deleted
                    self.model.dont_change_active = False
                self.model.delete_row_by_handle(hndl)
            LOG.debug(
                "   "
                + self.__class__.__name__
                + " row_delete "
                + str(perf_counter() - cput)
                + " sec"
            )
            if self.active:
                self.uistate.show_filter_results(
                    self.dbstate, self.model.displayed(), self.model.total()
                )
        else:
            self.dirty = True

    def object_build(self, *args):
        """
        Called when the tree must be rebuilt and bookmarks redrawn.
        """
        self.dirty = True
        if self.active:
            # Save the currently selected handles, if any:
            selected_ids = self.selected_handles()
            self.bookmarks.redraw()
            self.build_tree()
            # Reselect one, if it still exists after rebuild:
            nav_type = self.navigation_type()
            lookup_handle = self.dbstate.db.method("get_%s_from_handle", nav_type)
            for handle in selected_ids:
                # Still exist?
                try:
                    lookup_handle(handle)
                    # Select it, and stop selecting:
                except HandleError:
                    continue
                self.change_active(handle)
                break

    def related_update(self, hndl_list):
        """Find handles pointing to the view from a related object update;
        for example if an event update occurs, find person handles referenced
        by that event. Use the created list to perfom row_updates.
        Places need a bit more work, as they could be enclosing other places.
        In addition, for People view the birth/death place name could change.
        So we recursively check places and events until we find our class
        object handle to use for updating rows.
        """
        nav_type = self.navigation_type()
        upd_list = []
        done = set()
        queue = deque(hndl_list)
        while queue:
            hndl = queue.pop()
            if hndl in done:  # make sure we aren't in infinite loop
                continue  # in case places can enclose each other
            done.add(hndl)
            for cl_name, handle in self.dbstate.db.find_backlink_handles(hndl):
                if cl_name == nav_type:
                    upd_list.append(handle)
                    if len(upd_list) > 20:
                        self.dirty = True
                        return
                if cl_name == "Place" or cl_name == "Event" and nav_type == "Person":
                    queue.append(handle)
        if upd_list:
            self.row_update(upd_list)

    def _button_press(self, obj, event):
        """
        Called when a mouse is clicked.
        """
        if not self.dbstate.is_open():
            return False
        menu = self.uimanager.get_widget("Popup")
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            if self.model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY:
                self.edit(obj)
                return True
            else:
                # Tree
                store, paths = self.selection.get_selected_rows()
                if paths:
                    firstsel = self.model.get_iter(paths[0])
                    handle = self.model.get_handle_from_iter(firstsel)
                    if len(paths) == 1 and handle is None:
                        return self.expand_collapse_tree_branch()
                    else:
                        self.edit(obj)
                        return True
        elif is_right_click(event) and menu:
            prefix = "win"
            self.at_popup_menu = []
            actions = []
            # Quick Reports
            if self.QR_CATEGORY != CATEGORY_QR_MISC:
                (qr_ui, qr_actions) = create_quickreport_menu(
                    self.QR_CATEGORY,
                    self.dbstate,
                    self.uistate,
                    self.first_selected(),
                    prefix,
                )
                if self.get_active() and qr_actions:
                    actions.extend(qr_actions)
                    qr_ui = "<placeholder id='QuickReport'>%s</placeholder>" % qr_ui
                    self.at_popup_menu.append(qr_ui)

            # Web Connects
            if self.QR_CATEGORY != CATEGORY_QR_MISC:
                (web_ui, web_actions) = create_web_connect_menu(
                    self.dbstate,
                    self.uistate,
                    self.navigation_type(),
                    self.first_selected(),
                    prefix,
                )
                if self.get_active() and web_actions:
                    actions.extend(web_actions)
                    self.at_popup_menu.append(web_ui)

            if self.at_popup_action:
                self.uimanager.remove_ui(self.at_popup_menu)
                self.uimanager.remove_action_group(self.at_popup_action)
            self.at_popup_action = ActionGroup("AtPopupActions", actions)
            self.uimanager.insert_action_group(self.at_popup_action)
            self.at_popup_menu = self.uimanager.add_ui_from_string(self.at_popup_menu)
            self.uimanager.update_menu()

            menu = self.uimanager.get_widget("Popup")
            popup_menu = Gtk.Menu.new_from_model(menu)
            popup_menu.attach_to_widget(obj, None)
            popup_menu.show_all()
            popup_menu.popup_at_pointer(event)
            return True

        return False

    def _key_press(self, obj, event):
        """
        Called when a key is pressed on a listview
        """
        if not self.dbstate.is_open():
            return False
        if self.model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY:
            # Flat list
            return self._key_press_flat(obj, event)
        else:
            # Tree
            return self._key_press_tree(obj, event)

    def _key_press_flat(self, obj, event):
        """
        Called when a key is pressed on a flat listview
        ENTER --> edit selection
        """
        if event.keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter):
            self.edit(obj)
            return True
        # Custom interactive search
        if Gdk.keyval_to_unicode(event.keyval):
            return self.searchbox.treeview_keypress(obj, event)
        return False

    def _key_press_tree(self, obj, event):
        """
        Called when a key is pressed on a tree listview
        ENTER --> edit selection or open group node
        SHIFT+ENTER --> open group node and all children nodes
        """
        if event.get_state() & Gdk.ModifierType.SHIFT_MASK and event.keyval in (
            Gdk.KEY_Return,
            Gdk.KEY_KP_Enter,
        ):
            store, paths = self.selection.get_selected_rows()
            if paths:
                iter_ = self.model.get_iter(paths[0])
                handle = self.model.get_handle_from_iter(iter_)
                if len(paths) == 1 and handle is None:
                    return self.expand_collapse_tree_branch()
        elif event.keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter):
            store, paths = self.selection.get_selected_rows()
            if paths:
                iter_ = self.model.get_iter(paths[0])
                handle = self.model.get_handle_from_iter(iter_)
                if len(paths) == 1 and handle is None:
                    return self.expand_collapse_tree()
                else:
                    self.edit(obj)
                    return True
        elif Gdk.keyval_to_unicode(event.keyval):
            # Custom interactive search
            return self.searchbox.treeview_keypress(obj, event)
        return False

    def expand_collapse_tree(self):
        """
        Expand or collapse the selected group node.
        Return True if change done, False otherwise
        """
        store, paths = self.selection.get_selected_rows()
        if paths:
            firstsel = paths[0]
            iter_ = self.model.get_iter(firstsel)
            handle = self.model.get_handle_from_iter(iter_)
            if handle:
                return False
            if self.list.row_expanded(firstsel):
                self.list.collapse_row(firstsel)
            else:
                self.list.expand_row(firstsel, False)
            return True
        return False

    def expand_collapse_tree_branch(self):
        """
        Expand or collapse the selected group node with all children.
        Return True if change done, False otherwise
        """
        store, paths = self.selection.get_selected_rows()
        if paths:
            firstsel = paths[0]
            iter_ = self.model.get_iter(firstsel)
            handle = self.model.get_handle_from_iter(iter_)
            if handle:
                return False
            if self.list.row_expanded(firstsel):
                self.list.collapse_row(firstsel)
            else:
                self.open_branch(None)
            return True
        return False

    def change_page(self):
        """
        Called when a page is changed.
        """
        NavigationView.change_page(self)
        if self.model:
            self.uistate.show_filter_results(
                self.dbstate, self.model.displayed(), self.model.total()
            )
        self.uimanager.set_actions_visible(self.edit_action, True)
        self.uimanager.set_actions_sensitive(
            self.edit_action, not self.dbstate.db.readonly
        )

    def on_delete(self):
        """
        Save the column widths when the view is shutdown.
        """
        self.list.save_column_info()
        # The following is used when we change the columns order.
        self.save_column_info()
        PageView.on_delete(self)

    def get_config_name(self):
        """
        Set the associated config name string for the treeview
        """
        assert False, "Must be defined in the subclass"

    def save_column_info(self):
        """
        Save the column widths, order, and view settings
        This is used only when we change the columns order
        """
        widths = self.get_column_widths()
        order = self._config.get("columns.rank")
        size = self._config.get("columns.size")
        vis = self._config.get("columns.visible")
        newsize = []
        index = 0
        for val, size in zip(order, size):
            if val in vis[:-1]:  # don't use last column size, it's wrong
                if widths[index]:
                    size = widths[index]
                index += 1
            newsize.append(size)
        self._config.set("columns.size", newsize)

    ####################################################################
    # Export data
    ####################################################################
    def export(self, *obj):
        chooser = Gtk.FileChooserDialog(
            title=_("Export View as Spreadsheet"),
            transient_for=self.uistate.window,
            action=Gtk.FileChooserAction.SAVE,
        )
        chooser.add_buttons(
            _("_Cancel"), Gtk.ResponseType.CANCEL, _("_Save"), Gtk.ResponseType.OK
        )
        chooser.set_do_overwrite_confirmation(True)

        combobox = Gtk.ComboBoxText()
        label = Gtk.Label(label=_("Format:"))
        label.set_halign(Gtk.Align.END)
        box = Gtk.Box()
        box.pack_start(label, True, True, padding=12)
        box.pack_start(combobox, False, False, 0)
        combobox.append_text(_("CSV"))
        combobox.append_text(_("OpenDocument Spreadsheet"))
        combobox.set_active(0)
        box.show_all()
        chooser.set_extra_widget(box)
        default_dir = config.get("paths.recent-export-dir")
        chooser.set_current_folder(default_dir)

        while True:
            value = chooser.run()
            fn = chooser.get_filename()
            fl = combobox.get_active()
            if value == Gtk.ResponseType.OK:
                if fn:
                    chooser.destroy()
                    break
            else:
                chooser.destroy()
                return
        config.set("paths.recent-export-dir", os.path.split(fn)[0])
        self.write_tabbed_file(fn, fl)

    def write_tabbed_file(self, name, type):
        """
        Write a tabbed file to the specified name.

        The output file type is determined by the type variable.
        """
        from gramps.gen.utils.docgen import CSVTab, ODSTab

        ofile = None
        data_cols = [pair[1] for pair in self.column_order() if pair[0]]

        column_names = [self.COLUMNS[i][0] for i in data_cols]
        if type == 0:
            ofile = CSVTab(len(column_names))
        else:
            ofile = ODSTab(len(column_names))

        ofile.open(name)
        ofile.start_page()
        ofile.start_row()

        # Headings
        if self.model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY:
            headings = column_names
        else:
            levels = self.model.get_tree_levels()
            headings = levels + column_names[1:]
            data_cols = data_cols[1:]

        list(map(ofile.write_cell, headings))
        ofile.end_row()

        if self.model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY:
            # Flat model
            for row in self.model:
                ofile.start_row()
                for index in data_cols:
                    ofile.write_cell(row[index])
                ofile.end_row()
        else:
            # Tree model
            iter_ = self.model.get_iter((0,))
            if iter_:
                self.write_node(iter_, len(levels), [], ofile, data_cols)

        ofile.end_page()
        ofile.close()

    def write_node(self, iter_, depth, level, ofile, data_cols):
        while iter_:
            new_level = level + [self.model.get_value(iter_, 0)]
            if self.model.get_handle_from_iter(iter_):
                ofile.start_row()
                padded_level = new_level + [""] * (depth - len(new_level))
                list(map(ofile.write_cell, padded_level))
                for index in data_cols:
                    ofile.write_cell(self.model.get_value(iter_, index))
                ofile.end_row()

            first_child = self.model.iter_children(iter_)
            self.write_node(first_child, depth, new_level, ofile, data_cols)

            iter_ = self.model.iter_next(iter_)

    ####################################################################
    # Template functions
    ####################################################################
    @abstractmethod
    def edit(self, *obj):
        """
        Template function to allow the editing of the selected object
        """

    @abstractmethod
    def remove(self, *obj):
        """
        Template function to allow the removal of an object by its handle
        """

    @abstractmethod
    def add(self, *obj):
        """
        Template function to allow the adding of a new object
        """

    @abstractmethod
    def merge(self, *obj):
        """
        Template function to allow the merger of two objects.
        """

    def open_all_nodes(self, *obj):
        """
        Method for Treeviews to open all groups
        obj: for use of method in event callback
        """
        self.uistate.status_text(_("Updating display..."))
        self.uistate.set_busy_cursor(True)

        self.list.expand_all()

        self.uistate.set_busy_cursor(False)
        self.uistate.modify_statusbar(self.dbstate)

    def close_all_nodes(self, *obj):
        """
        Method for Treeviews to close all groups
        obj: for use of method in event callback
        """
        self.list.collapse_all()

    def open_branch(self, *obj):
        """
        Expand the selected branches and all children.
        obj: for use of method in event callback
        """
        self.uistate.status_text(_("Updating display..."))
        self.uistate.set_busy_cursor(True)

        store, selected = self.selection.get_selected_rows()
        for path in selected:
            self.list.expand_row(path, False)

        self.uistate.set_busy_cursor(False)
        self.uistate.modify_statusbar(self.dbstate)

    def close_branch(self, *obj):
        """
        Collapse the selected branches.
        :param obj: not used, present only to allow the use of the method in
            event callback
        """
        store, selected = self.selection.get_selected_rows()
        for path in selected:
            self.list.collapse_row(path)

    def can_configure(self):
        """
        See :class:`~gui.views.pageview.PageView
        :return: bool
        """
        return True

    def config_connect(self):
        """
        Overwriten from  :class:`~gui.views.pageview.PageView method
        This method will be called after the ini file is initialized,
        use it to monitor changes in the ini file
        """
        # func = self.config_callback(self.build_tree)
        # self._config.connect('columns.visible', func)
        # self._config.connect('columns.rank', func)
        pass

    def _get_configure_page_funcs(self):
        """
        Return a list of functions that create gtk elements to use in the
        notebook pages of the Configure dialog

        :return: list of functions
        """

        def columnpage(configdialog):
            flat = self.model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY
            column_names = [col[0] for col in self.COLUMNS]
            return _("Columns"), ColumnOrder(
                self._config,
                column_names,
                self.get_column_widths(),
                self.set_column_order,
                tree=not flat,
                resizable=self.list,
            )

        def csvdialect(configdialog):
            return _("CSV Dialect"), CsvDialect()

        return [columnpage, csvdialect]



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# plugins/rel/rel_hr.py
#
# Copyright (C) 2010 Josip (josip at pisoj dot com)

"""
Croatian-specific classes for calculating relationships and kinship names.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

import gramps.gen.relationship

_PARENTS = [
    "",
    "otac i majka",
    "djedovi i bake",
    "pradjedovi i prabake",
    "%s pradjedovi i %s prabake",
]
_CHILDS = ["", "djeca", "unučad", "praunučad", "%s praunučad"]
_CHILDS_SP = [
    "",
    "snahe i zetovi",
    "prasnahe i prazetovi",
    "supružnici praunučadi",
    "supružnici %s praunučadi",
]


# -------------------------------------------------------------------------
#
# Class HrDeclination
#
# -------------------------------------------------------------------------
class HrDeclination:
    """
    Croatian declination system
    """

    numdcl = {
        1: "prv",
        2: "drug",
        3: "treć",
        4: "četvrt",
        5: "pet",
        6: "šest",
        7: "sedm",
        8: "osm",
        9: "devet",
        10: "deset",
        11: "jedanaest",
        12: "dvanaest",
        13: "trinaest",
        14: "četrnaest",
        15: "petnaest",
        16: "šestnaest",
        17: "sedamnaest",
        18: "osamnaest",
        19: "devetnaest",
        20: "dvadeset",
        30: "trideset",
        40: "četrdeset",
        50: "pedeset",
        60: "šesdeset",
        70: "sedamdeset",
        80: "osamdeset",
        90: "devedeset",
        (1, 1, 1): "i",
        (1, 1, 5): "i",
        (1, 2, 1): "i",
        (1, 2, 5): "i",
        (1, 1, 2): "og",
        (1, 1, 4): "og",
        (2, 1, 2): "og",
        (1, 1, 3): "om",
        (1, 1, 6): "om",
        (0, 1, 7): "om",
        (2, 1, 3): "om",
        (2, 1, 6): "om",
        (2, 2, 7): "om",
        (1, 1, 7): "im",
        (2, 1, 7): "im",
        (1, 2, 3): "im",
        (1, 2, 6): "im",
        (1, 2, 7): "im",
        (0, 2, 3): "im",
        (0, 2, 6): "im",
        (0, 2, 7): "im",
        (0, 1, 1): "a",
        (0, 1, 5): "a",
        (2, 2, 1): "a",
        (2, 2, 5): "a",
        (0, 1, 2): "e",
        (0, 1, 4): "e",
        (1, 2, 4): "e",
        (0, 2, 1): "e",
        (0, 2, 4): "e",
        (0, 2, 5): "e",
        (2, 1, 1): "o",
        (2, 1, 4): "o",
        (2, 1, 5): "o",
        (0, 1, 3): "oj",
        (0, 1, 6): "oj",
        (2, 2, 3): "oj",
        (2, 2, 6): "oj",
        (0, 2, 2): "ih",
        (1, 2, 2): "ih",
        (2, 2, 2): "ih",
        (2, 2, 4): "u",
    }

    def get_ordnum(self, num, gender, number, case):
        """
        Declination of ordinal numbers
        gender:   0 = feminine, 1 = masculine, 2 = neuter
        case:       1 = nominative, 2 = genitive, 3 = dative, 4 = accusative
                        5 = vocative, 6 = locative, 7 = instrumental
        number:   1 = single, 2 = plural
        :rtype: str
        """
        if 0 < num < 100:
            lres = []
            # lres.append(self.numdcl[(num / 100) * 100])
            if (num < 21) or (num % 10) == 0:
                lres.append(self.numdcl[num])
            else:
                lres.append(self.numdcl[(num % 10) * 10])
                lres.append(self.numdcl[(num % 10)])
            sufix = self.numdcl[(gender, number, case)]
            if num % 10 == 3:
                if gender != 0 and sufix != "oj":
                    sufix = sufix.replace("o", "e")
            lres[-1] += sufix
            res = " ".join(lres)
        else:
            res = str(num) + ". "
        return res


HRD = HrDeclination()
HRDN = HRD.get_ordnum


def _get_childs(level, inlaw):
    """children, grandchildren et."""
    if level < len(_CHILDS) - 1:
        if not inlaw:
            return _CHILDS[level]
        else:
            return _CHILDS_SP[level]
    else:
        if not inlaw:
            return "%s" % _CHILDS[4] % (HRDN(level - 2, 0, 1, 1))
        elif inlaw:
            return "%s" % _CHILDS_SP[4] % (HRDN(level - 2, 0, 2, 1))


def _get_parents(level):
    """parents, grandparents"""
    if level < 4:
        return _PARENTS[level]
    else:
        return "%s" % _PARENTS[4] % (HRDN(level - 2, 1, 1, 1), HRDN(level - 2, 0, 2, 1))


def _get_uncles(gen, inlaw):
    """in general: uncles, ants---"""
    if gen == 2:
        if not inlaw:
            return "stričevi, ujaci i tetke"
        else:
            return "strine, ujne i tetci"
    elif gen == 3:
        if not inlaw:
            return "prastričevi, praujaci i pratetke"
        else:
            return "prastrine, praujne i pratetci"
    else:
        if not inlaw:
            return "%s prastričevi, %s praujaci i %s pratetke" % (
                HRDN(gen - 2, 1, 2, 1),
                HRDN(gen - 2, 1, 2, 1),
                HRDN(gen - 2, 0, 2, 1),
            )
        else:
            return "prastrine, praujne i pratetci"


# -------------------------------------------------------------------------
#
#  CroatianRelationshipCalculator
#
# -------------------------------------------------------------------------


class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_plural_relationship_string(
        self,
        gena,
        genb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """returns kinship terms for groups of persons"""
        if not in_law_b:
            rel_str = "potomci %s generacije predaka u %s koljenu: " % (
                HRDN(gena, 0, 1, 2),
                HRDN(genb, 2, 1, 6),
            )
        elif in_law_b:
            rel_str = "supružnici potomaka %s generacije predaka u %s " "koljenu: " % (
                HRDN(gena, 0, 1, 2),
                HRDN(genb, 2, 1, 6),
            )
        if genb == 0:
            # These are ancestors
            rel_str = "%s generacija predaka: %s" % (
                HRDN(gena, 0, 1, 1),
                _get_parents(gena),
            )
        elif gena == 0:
            # These are descendants
            if not in_law_b:
                rel_str = "%s generacija potomaka: %s" % (
                    HRDN(genb, 0, 1, 1),
                    _get_childs(genb, in_law_b),
                )
            else:
                rel_str = "supružnici %s generacije potomaka: %s" % (
                    HRDN(genb, 0, 2, 1),
                    _get_childs(genb, in_law_b),
                )
        elif gena == 1 == genb:
            # These are siblings
            if not in_law_b:
                rel_str += "braća i sestre"
            else:
                rel_str = "supružnici braće i sestara"
        elif genb == 1:
            # These are aunts/uncles
            rel_str += _get_uncles(gena, in_law_b)
        return rel_str


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    # python src/plugins/rel/rel_ru.py
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    SurnameListPage - Index for first letters of surname
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from decimal import getcontext
import logging
from collections import defaultdict
from unicodedata import name as uniname

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.gen.display.name import displayer as _nd
from gramps.plugins.webreport.common import (
    alphabet_navigation,
    html_escape,
    name_to_md5,
    FULLCLEAR,
    get_surname_from_person,
    AlphabeticIndex,
)

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


#################################################
#
#    Creates the Surname List page
#
#################################################
class SurnameListPage(BasePage):
    """
    This class is responsible for displaying the list of Surnames
    """

    ORDER_BY_NAME = 0
    ORDER_BY_COUNT = 1

    def __init__(
        self,
        report,
        the_lang,
        the_title,
        ppl_handle_list,
        order_by=ORDER_BY_NAME,
        filename="surnames",
    ):
        """
        @param: report          -- The instance of the main report class for
                                   this report
        @param: the_lang        -- The lang to process
        @param: the_title       -- The title page related to the language
        @param: ppl_handle_list -- The list of people for whom we need to create
                                   a page.
        @param: order_by        -- The way to sort surnames :
                                   Surnames or Surnames count
        @param: filename        -- The name to use for the Surnames page
        """
        BasePage.__init__(self, report, the_lang, the_title)

        if order_by == self.ORDER_BY_NAME:
            output_file, sio = self.report.create_file(filename)
            result = self.write_header(self._("Surnames"))
        else:
            output_file, sio = self.report.create_file("surnames_count")
            result = self.write_header(self._("Surnames by person count"))
        surnamelistpage, dummy_head, dummy_body, outerwrapper = result

        # begin surnames division
        with Html("div", class_="content", id="surnames") as surnamelist:
            outerwrapper += surnamelist

            # page message
            msg = self._(
                "This page contains an index of all the "
                "surnames in the database. Selecting a link "
                "will lead to a list of individuals in the "
                "database with this same surname."
            )
            surnamelist += Html("p", msg, id="description")

            # Assemble all the handles for each surname into a dictionary
            # We don't call sort_people because we don't care about sorting
            # individuals, only surnames
            surname_handle_dict = defaultdict(list)
            for person_handle in ppl_handle_list:
                person = self.r_db.get_person_from_handle(person_handle)
                surname = get_surname_from_person(self.r_db, person)
                surname_handle_dict[surname].append(person_handle)

            # add alphabet navigation...
            # only if surname list not surname count
            if order_by == self.ORDER_BY_NAME:
                # Assemble the AlphabeticIndex
                index = AlphabeticIndex(self.rlocale)
                for surname, handle_list in surname_handle_dict.items():
                    index.addRecord(surname, handle_list)

                # Extract the buckets from the index
                index_list = []
                index.resetBucketIterator()
                while index.nextBucket():
                    if index.bucketRecordCount != 0:
                        index_list.append(index.bucketLabel)
                # Output the navigation
                alpha_nav = alphabet_navigation(index_list, self.rlocale, rtl=self.dir)
                if alpha_nav is not None:
                    surnamelist += alpha_nav

            if order_by == self.ORDER_BY_COUNT:
                table_id = "SortByCount"
            else:
                table_id = "SortByName"

            # begin surnamelist table and table head
            with Html(
                "table",
                class_="infolist primobjlist surnamelist " + self.dir,
                id=table_id,
            ) as table:
                surnamelist += table

                thead = Html("thead")
                table += thead

                trow = Html("tr")
                thead += trow

                trow += Html("th", self._("Letter"), class_="ColumnLetter", inline=True)

                # create table header surname hyperlink
                fname = self.report.surname_fname + self.ext
                tcell = Html("th", class_="ColumnSurname", inline=True)
                trow += tcell
                hyper = Html(
                    "a", self._("Surname"), href=fname, title=self._("Surnames")
                )
                tcell += hyper

                # create table header number of people hyperlink
                fname = "surnames_count" + self.ext
                tcell = Html("th", class_="ColumnQuantity", inline=True)
                trow += tcell
                num_people = self._("Number of People")
                hyper = Html("a", num_people, href=fname, title=num_people)
                tcell += hyper

                name_format = self.report.options["name_format"]
                # begin table body
                with Html("tbody") as tbody:
                    table += tbody

                    if order_by == self.ORDER_BY_COUNT:
                        # construct a dictionary of counts, for example
                        # {3: ["brown", [handle1, handle2,handle3],
                        #      "smith", [handle4, handle5, handle6]
                        #     ],
                        #  2: ["Jones", [handle7]
                        #
                        # }]
                        count_ppl_handle_dict = defaultdict(list)
                        for surname, data_list in surname_handle_dict.items():
                            count_ppl_handle_dict[len(data_list)].append(
                                (surname, data_list)
                            )
                        # For each count, we construct and output a separate
                        # AlphabeticIndex for all surnames with that count
                        for dummy_count, ppl_handles in sorted(
                            count_ppl_handle_dict.items(), reverse=True
                        ):
                            # Construct the AplhabeticIndex for that count
                            index = AlphabeticIndex(self.rlocale)
                            for surname, handle_list in ppl_handles:
                                index.addRecord(surname, handle_list)
                            # Output the AlphabeticIndex for that count
                            self.output_surname_records(index, tbody, name_format)

                    else:  # order_by == self.ORDER_BY_NAME
                        # The AlphabeticIndex has already been constructed
                        # Output the AlphabeticIndex
                        self.output_surname_records(index, tbody, name_format)

        # create footer section
        # add clearline for proper styling
        footer = self.write_footer(None)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(
            surnamelistpage, output_file, sio, 0
        )  # 0 => current date modification

    def surname_link(self, fname, name, opt_val=None, uplink=False):
        """
        Create a link to the surname page.

        @param: fname   -- Path to the file name
        @param: name    -- Name to see in the link
        @param: opt_val -- Option value to use
        @param: uplink  -- If True, then "../../../" is inserted in front of
                           the result.
        """
        url = self.report.build_url_fname_html(fname, "srn", uplink)
        try:  # some characters don't have a unicode name
            char = uniname(name[0])
        except (ValueError, TypeError) as dummy_err:
            char = " "
        hyper = Html(
            "a",
            html_escape(name),
            href=url,
            title="%s starting with %s" % (name, char),
            inline=True,
        )
        if opt_val is not None:
            hyper += opt_val

        # return hyperlink to its caller
        return hyper

    def output_surname_records(self, index, tbody, name_format):
        """
        Output all the surnames in the index.

        @param: index   -- An ICU AlphabeticIndex where the names are surnames
                           and the data is a list of people handles with that
                           surname
        @param: tbody   -- The HTML body to which the lines are added
        @param: name_format -- The name format from the report options
        """
        index.resetBucketIterator()
        output = []
        dup_index = 0
        while index.nextBucket():
            if index.bucketRecordCount != 0:
                trow = Html("tr")
                tbody += trow

                tcell = Html("td", class_="ColumnLetter", inline=True)
                trow += tcell
                trow.attr = 'class = "BeginLetter"'
                bucket_letter = index.bucketLabel
                bucket_link = bucket_letter
                if bucket_letter in output:
                    bucket_link = "%s (%i)" % (bucket_letter, dup_index)
                    dup_index += 1
                output.append(bucket_letter)
                try:  # some characters don't have a unicode name
                    char = uniname(bucket_letter)
                except (ValueError, TypeError) as dummy_err:
                    char = " "
                ttle = self._(
                    "Surnames beginning with letter '{letter}' {char}"
                ).format(
                    letter=bucket_letter,
                    char=char,
                )
                hyper = Html(
                    "a",
                    index.bucketLabel,
                    name=index.bucketLabel,
                    id_=bucket_link,
                    title=ttle,
                )
                tcell += hyper

                first = True
                while index.nextRecord():
                    surname = index.recordName
                    if not first:
                        trow = Html("tr")
                        tbody += trow

                        tcell = Html("td", class_="ColumnLetter", inline=True)
                        trow += tcell
                    first = False
                    if not surname or surname.isspace():
                        surname = self._("<absent>")
                    # In case the user choose a format name like "*SURNAME*"
                    # We must display this field in upper case. So we use
                    # the english format of format_name to find if this is
                    # the case.
                    # name_format = self.report.options['name_format']
                    nme_format = _nd.name_formats[name_format][1]
                    if "SURNAME" in nme_format:
                        surnamed = surname.upper()
                    else:
                        surnamed = surname
                    trow += Html(
                        "td",
                        self.surname_link(name_to_md5(surname), surnamed),
                        class_="ColumnSurname",
                        inline=True,
                    )

                    trow += Html(
                        "td",
                        len(index.recordData),
                        class_="ColumnQuantity",
                        inline=True,
                    )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# test/test/test_util_test.py

"""unittest (test_util_test.py) for test_util.py"""
import sys
import os
import tempfile
import unittest as U

from gramps.gen.constfunc import get_env_var

usage_note = """
  **************************************************************
  Testing (and runing) Gramps requires that PYTHONPATH include
  the path to the top Gramps directory (where gramps.py resides).

  For example, in bash, a shell export would look like
    export PYTHONPATH=/.../src
  with the ... filled in appropriately.
  **************************************************************
"""

# **************************************************************
#
# Since this module is used by other test modules, it is
# strongly advised to test this module to 100% coverage,
# and in all calling variations, eg:
#  run directly, from this dir with and without ./ prefix
#  run from other dirs (with path prefix)
#  run from within regrtest.py
#  run from regrtest.py with other test modules present
#    which use the test_util module itself
#
# **************************************************************

try:
    from gramps.test import test_util as tu

    ##here = tu.absdir()
except ImportError:
    print("Cannot import 'test_util' from package 'test'" + usage_note)
    exit(1)


# grouping into multiple TestCases (classes) is not required,
# but may be useful for various reasons, such as collecting
# tests that share a setUp/tearDown mechanism or that share
# some test data, or just because they're related.
#
# The test function name should not have docstrings, but should
# have names which add to the value of failure reporting, and
# which make it easy to find them within the source.


# some enabling infrastructure features
class Test1(U.TestCase):
    def test1a_custom_exception(self):
        tmsg = "testing"
        try:
            err = None
            raise tu.TestError(tmsg)
        except tu.TestError as e:
            emsg = e.value
        self.assertEqual(emsg, tmsg, "raising TestError: g=%r e=%r" % (emsg, tmsg))

    def test1b_msg_reporting_utility(self):
        g, e = "got this", "expected that"
        m, p = "your message here", "pfx"
        tmsg0 = m + "\n .....got:'" + g + "'\n expected:'" + e + "'"
        tmsg1 = p + ": " + tmsg0
        self.assertEqual(tu.msg(g, e, m), tmsg0, "non-prefix message")
        self.assertEqual(tu.msg(g, e, m, p), tmsg1, "prefix message")


# path-related features (note use of tu.msg tested above)
class Test2(U.TestCase):
    def test2a_context_via_traceback(self):
        e = os.path.basename(__file__).rstrip(".co")  # eg in *.py[co]
        g = os.path.basename(tu._caller_context()[0]).rstrip("co")
        self.assertEqual(g, e, tu.msg(g, e, "_caller_context"))

    def test2b_absdir(self):
        here = tu.absdir()
        g = tu.absdir(__file__)
        self.assertEqual(g, here, tu.msg(g, here, "absdir"))

    def test2c_path_append_parent(self):
        here = tu.absdir()
        par = os.path.dirname(here)
        was_there = par in sys.path
        if was_there:
            while par in sys.path:
                sys.path.remove(par)
        np = len(sys.path)

        for p in (None, __file__):
            self.assertFalse(par in sys.path, "par not in initial path")
            if not p:
                g = tu.path_append_parent()
            else:
                g = tu.path_append_parent(p)
            self.assertEqual(g, par, tu.msg(g, par, "path_append_parent return"))
            self.assertTrue(par in sys.path, "actually appends")
            sys.path.remove(par)
            l = len(sys.path)
            self.assertEqual(l, np, tu.msg(l, np, "numpaths"))
        if was_there:
            # restore entry state (but no multiples needed!)
            sys.path.append(par)


# make and remove test dirs
class Test3(U.TestCase):
    here = tu.absdir()
    bases = (here, tempfile.gettempdir())
    asubs = [os.path.join(b, "test_sub") for b in bases]
    home = get_env_var("HOME")
    if home:
        home_junk = os.path.join(home, "test_junk")

    def _rmsubs(self):
        import shutil

        for sub in self.asubs:
            if os.path.isdir(sub):
                shutil.rmtree(sub)

    def setUp(self):
        self._rmsubs()
        if self.home and not os.path.isdir(self.home_junk):
            os.mkdir(self.home_junk)

    def tearDown(self):
        self._rmsubs()
        if self.home and os.path.isdir(self.home_junk):
            os.rmdir(self.home_junk)

    def test3a_subdir(self):
        for sub in self.asubs:
            self.assertFalse(os.path.isdir(sub), "init: no dir %r" % sub)
            b, d = os.path.dirname(sub), os.path.basename(sub)
            md = tu.make_subdir(d, b)
            self.assertTrue(os.path.isdir(sub), "made dir %r" % sub)
            self.assertEqual(md, sub, tu.msg(md, sub, "make_subdir returns path"))

            s2 = os.path.join(sub, "sub2")
            tu.make_subdir("sub2", sub)
            self.assertTrue(os.path.isdir(s2), "made dir %r" % s2)
            f = os.path.join(s2, "test_file")

            with open(f, "w") as fd:
                fd.write("testing..")
            self.assertTrue(os.path.isfile(f), "file %r exists" % f)
            tu.delete_tree(sub)
            self.assertFalse(os.path.isdir(sub), "delete_tree removes subdir %r" % sub)

    def test3b_delete_tree_constraint(self):
        if self.home:
            sep = os.path.sep
            tmp = tempfile.gettempdir() + sep
            if (self.home + sep).startswith(tmp):
                self.skipTest("skipping delete_tree_constraint test")
            err = None
            try:
                tu.delete_tree(self.home_junk)
            except tu.TestError as e:
                err = e.value
            self.assertFalse(
                err is None, "deltree on %r raises TestError" % (self.home_junk)
            )
        else:
            self.fail("Skip deltree constraint test, no '$HOME' var")


if __name__ == "__main__":
    U.main()
# ===eof===



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006  Donald N. Allingham
# Copyright (C) 2008,2012  Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2012       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""custom text for the book report"""

# Written by Alex Roitman,
# largely based on the SimpleBookTitle.py by Don Allingham

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# gtk
#
# ------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.plug.menu import TextOption
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.docgen import (
    FontStyle,
    ParagraphStyle,
    FONT_SANS_SERIF,
    PARA_ALIGN_CENTER,
    IndexMark,
    INDEX_TYPE_TOC,
)


# ------------------------------------------------------------------------
#
# CustomText
#
# ------------------------------------------------------------------------
class CustomText(Report):
    """CustomText"""

    def __init__(self, database, options, user):
        """
        Create CustomText object that produces the report.

        The arguments are:

        database        - the Gramps database instance
        options         - instance of the Options class for this report
        user            - a gen.user.User() instance

        This report needs the following parameters (class variables)
        that come in the options class.

        top   - Text on the top.
        mid   - Text in the middle.
        bot   - Text on the bottom.
        """
        Report.__init__(self, database, options, user)

        menu = options.menu
        self.top_text = menu.get_option_by_name("top").get_value()
        self.middle_text = menu.get_option_by_name("mid").get_value()
        self.bottom_text = menu.get_option_by_name("bot").get_value()

    def write_report(self):
        mark_text = _("Custom Text")
        if self.top_text[0]:
            mark_text = "%s (%s)" % (_("Custom Text"), self.top_text[0])
        elif self.middle_text[0]:
            mark_text = "%s (%s)" % (_("Custom Text"), self.middle_text[0])
        elif self.bottom_text[0]:
            mark_text = "%s (%s)" % (_("Custom Text"), self.bottom_text[0])
        mark = IndexMark(mark_text, INDEX_TYPE_TOC, 1)
        self.doc.start_paragraph("CBT-Initial")
        for line in self.top_text:
            self.doc.write_text(line, mark)
            self.doc.write_text("\n")
        self.doc.end_paragraph()

        self.doc.start_paragraph("CBT-Middle")
        for line in self.middle_text:
            self.doc.write_text(line)
            self.doc.write_text("\n")
        self.doc.end_paragraph()

        self.doc.start_paragraph("CBT-Final")
        for line in self.bottom_text:
            self.doc.write_text(line)
            self.doc.write_text("\n")
        self.doc.end_paragraph()


# ------------------------------------------------------------------------
#
# CustomTextOptions
#
# ------------------------------------------------------------------------
class CustomTextOptions(MenuReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        self.__top = None
        self.__mid = None
        self.__bot = None
        MenuReportOptions.__init__(self, name, dbase)

    def add_menu_options(self, menu):
        category_name = _("Text")

        self.__top = TextOption(_("Initial Text"), [""])
        self.__top.set_help(_("Text to display at the top"))
        menu.add_option(category_name, "top", self.__top)

        self.__mid = TextOption(_("Middle Text"), [""])
        self.__mid.set_help(_("Text to display in the middle"))
        menu.add_option(category_name, "mid", self.__mid)

        self.__bot = TextOption(_("Final Text"), [""])
        self.__bot.set_help(_("Text to display at the bottom"))
        menu.add_option(category_name, "bot", self.__bot)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        if len(self.__top.get_value()[0]) > 0:
            return self.__top.get_value()[0]
        if len(self.__mid.get_value()[0]) > 0:
            return self.__mid.get_value()[0]
        if len(self.__bot.get_value()[0]) > 0:
            return self.__bot.get_value()[0]
        return ""

    def make_default_style(self, default_style):
        """Make the default output style for the Custom Text report."""
        font = FontStyle()
        font.set(face=FONT_SANS_SERIF, size=12, bold=0, italic=0)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set(pad=0.5)
        para.set_description(_("Text to display at the top"))
        default_style.add_paragraph_style("CBT-Initial", para)

        font = FontStyle()
        font.set(face=FONT_SANS_SERIF, size=12, bold=0, italic=0)
        para = ParagraphStyle()
        para.set_font(font)
        para.set(pad=0.5)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_description(_("Text to display in the middle"))
        default_style.add_paragraph_style("CBT-Middle", para)

        font = FontStyle()
        font.set(face=FONT_SANS_SERIF, size=12, bold=0, italic=0)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set(pad=0.5)
        para.set_description(_("Text to display at the bottom"))
        default_style.add_paragraph_style("CBT-Final", para)
